/******************************************************************************
*                                                                             *
*                      SOFTING Industrial Automation GmbH                     *
*                          Richard-Reitzner-Allee 6                           *
*                                D-85540 Haar                                 *
*                        Phone: (++49)-(0)89-45656-0                          *
*                        Fax:   (++49)-(0)89-45656-399                        *
*                                                                             *
*           Copyright (C) SOFTING Industrial Automation GmbH 2007-2010        *
*                              All Rights Reserved                            *
*                                                                             *
*******************************************************************************
FILE_NAME               PROFILE.C

PROJECT_NAME            ----

MODULE                  ----

COMPONENT_LIBRARY       ----

AUTHOR                  SOFTING

VERSION                 5.45.0.00.release

DATE                    30-November-2010

STATUS                  finished


FUNCTIONAL_MODULE_DESCRIPTION

The module is able to make a complete INI-file handling. It is intended to use
similiar to the respective functions '...PrivateProfile...' under MS-WINDOWS.

Use the module functions as follows:
- profile_open, profile_close: initializes the module for the duration of the whole
                               program; once memory has been allocated it will not be
                               released until termination

- profile_read               : the functions read the specified INI-file from
                               hard disk
- profile_get_string,
  profile_get_int            : these functions are the handlers for easy working with
                               INI-file entries


RELATED_DOCUMENTS
=============================================================================*/

#include "keywords.h"

INCLUDES


#include <stdio.h>

#ifdef WIN32
#include <windows.h>
#endif

#ifdef LINUX
#include <string.h>
#include <malloc.h>
#endif

#include "pb_type.h"

/****************************************************************************/

LOCAL_DEFINES

#define MAX_STRING_LEN      81           /* maximum length of local strings */
#define MAX_PROFILE_WIDTH   80                /* maximum length of INI line */
#define MAX_MEMORY_LINES    30        /* number of line elements per malloc */

#define REM                 ';'            /* comment character in INI file */

#define PROFILE_EMPTY       0                           /* type of INI line */
#define PROFILE_REM         1
#define PROFILE_LINE        2
#define PROFILE_GROUP       3

#define EOS               '\0'                              /* End Of String */
#define BS                0x08
#define HT                0x09
#define CR                0x0D
#define LF                0x0A
#define ESC               0x1B
#define SP                0x20
#define DEL               0x7F

/****************************************************************************/

LOCAL_TYPEDEFS

typedef struct T_MEMORY_DESC
{
  struct T_MEMORY_DESC  *next_ptr;
  VOID                  *memory_ptr;

} T_MEMORY_DESC;

/*---------------------------------------------------------------------------*/

typedef struct T_LINES_DESC
{
  struct T_LINES_DESC   *next_ptr,
                        *prev_ptr;

  char                  line [MAX_PROFILE_WIDTH + 1];
  size_t                line_length;
  unsigned char         line_type;

} T_LINES_DESC;

/****************************************************************************/

FUNCTION_DECLARATIONS

static BOOL allocate_memory_block (VOID);


/****************************************************************************/

EXPORT_DATA

/****************************************************************************/

IMPORT_DATA

/****************************************************************************/

LOCAL_DATA

static T_MEMORY_DESC    memory_desc_head;  /* anchor for memory description */

static T_LINES_DESC     used_lines_head,    /* doubled chained list anchors */
                        used_lines_tail,
                        free_lines_head,
                        free_lines_tail;


/****************************************************************************/



/****************************************************************************/

FUNCTION LOCAL BOOL check_ini_line
  (
    IN    T_LINES_DESC        *line_desc_ptr
  )
  /*------------------------------------------------------------------------*/
  /* FUNCTIONAL_DESCRIPTION                                                 */
  /*------------------------------------------------------------------------*/
  /* checks whether the current line is a valid INI-file line or not        */
  /*------------------------------------------------------------------------*/
{
  LOCAL_VARIABLES

  register char     *ini_line_ptr;
  register size_t    ini_line_length;

  register char      c, sep;
  unsigned char      type;


  FUNCTION_BODY

  ini_line_ptr    = line_desc_ptr->line;
  ini_line_length = line_desc_ptr->line_length;

  if (ini_line_length == 0) return (TRUE);

  if (*ini_line_ptr == '[')
  {
    ini_line_ptr++;
    ini_line_length--;
    sep  = ']';
    type = PROFILE_GROUP;
  }
  else
  {
    sep  = '=';
    type = PROFILE_LINE;
  }

  while (ini_line_length--)
  {
    c = *ini_line_ptr++;

    if (c == sep)
    {
      line_desc_ptr->line_type = type;
      return (TRUE);
    }

    if (c == REM)
    {
      line_desc_ptr->line_type = PROFILE_REM;
      return (TRUE);
    }
  } /* while */

  return (FALSE);
} /* check_ini_line */


/****************************************************************************/

FUNCTION LOCAL T_LINES_DESC *find_profile_group
  (
    IN    const char *group_string_ptr
  )
  /*------------------------------------------------------------------------*/
  /* FUNCTIONAL_DESCRIPTION                                                 */
  /*------------------------------------------------------------------------*/
  /* - looks for the via group_string_ptr specified group section within    */
  /*   the current INI-file                                                 */
  /* - returns the description structure pointer or NULL if fails           */
  /*------------------------------------------------------------------------*/
{
  LOCAL_VARIABLES

  register T_LINES_DESC  *line_desc_ptr;

  FUNCTION_BODY

  line_desc_ptr = used_lines_head.next_ptr;

  while (line_desc_ptr != line_desc_ptr->next_ptr)
  {
    if ( (line_desc_ptr->line_type == PROFILE_GROUP) &&
         (strncmp (line_desc_ptr->line + 1, group_string_ptr, strlen (group_string_ptr)) == 0) )
    {
      return (line_desc_ptr);
    }

    line_desc_ptr = line_desc_ptr->next_ptr;
  } /* while lines */

  return (NULL);
} /* find_profile_group */

/****************************************************************************/

FUNCTION LOCAL T_LINES_DESC     *find_profile_name
  (
    IN    T_LINES_DESC   *group_desc_ptr,
    IN    const char     *name_string_ptr
  )
  /*------------------------------------------------------------------------*/
  /* FUNCTIONAL_DESCRIPTION                                                 */
  /*------------------------------------------------------------------------*/
  /* - looks for the via name_string_ptr specified INI-file entry within    */
  /*   the via group_string_ptr specified group section                     */
  /* - returns the description structure pointer or NULL if fails           */
  /*------------------------------------------------------------------------*/
{
  LOCAL_VARIABLES

  register T_LINES_DESC  *line_desc_ptr;

  char                    str [MAX_STRING_LEN];
  register size_t         length;


  FUNCTION_BODY

  line_desc_ptr = group_desc_ptr->next_ptr;

  sprintf (str, "%s=", name_string_ptr);
  length = strlen (str);

  while ( (line_desc_ptr            != &used_lines_tail) &&
          (line_desc_ptr->line_type != PROFILE_GROUP   ) )
  {
    if ( (strncmp (line_desc_ptr->line, str, length) == 0           ) &&
         (line_desc_ptr->line_type                   == PROFILE_LINE) )
    {
      return (line_desc_ptr);
    }

    line_desc_ptr = line_desc_ptr->next_ptr;
  } /* while lines */

  return (NULL);
} /* find_profile_name */

/****************************************************************************/
/*** DYNAMIC MEMORY ACCESS **************************************************/
/****************************************************************************/

FUNCTION LOCAL VOID put_line_desc
  (
    IN    T_LINES_DESC   *line_desc_ptr,
    IN    T_LINES_DESC   *lower_ptr,
    IN    T_LINES_DESC   *upper_ptr
  )
  /*------------------------------------------------------------------------*/
  /* FUNCTIONAL_DESCRIPTION                                                 */
  /*------------------------------------------------------------------------*/
  /* - inserts the via line_desc_ptr specified INI-file entry between the   */
  /*   lower and upper elements                                             */
  /*------------------------------------------------------------------------*/
{
  LOCAL_VARIABLES

  FUNCTION_BODY

  line_desc_ptr->prev_ptr = upper_ptr->prev_ptr;
  upper_ptr    ->prev_ptr = line_desc_ptr;

  line_desc_ptr->next_ptr = lower_ptr->next_ptr;
  lower_ptr    ->next_ptr = line_desc_ptr;

} /* put_line_desc */

/****************************************************************************/

FUNCTION LOCAL T_LINES_DESC *get_line_desc
  (
    IN    T_LINES_DESC  *lower_ptr,
    IN    T_LINES_DESC  *upper_ptr
  )
  /*------------------------------------------------------------------------*/
  /* FUNCTIONAL_DESCRIPTION                                                 */
  /*------------------------------------------------------------------------*/
  /* - delivers the upper next INI-file entry if lower_ptr was chosen       */
  /* - delivers the lower next INI-file entry if upper_ptr was chosen       */
  /* - if there are no elements available NULL will be returned             */
  /*------------------------------------------------------------------------*/
{
  LOCAL_VARIABLES

  register T_LINES_DESC  *line_desc_ptr;

  FUNCTION_BODY

  if (lower_ptr != NULL)
  {
    line_desc_ptr = lower_ptr->next_ptr;

    if (line_desc_ptr == line_desc_ptr->next_ptr)
    {
      return (NULL);
    }
    else
    {
      lower_ptr->next_ptr               = line_desc_ptr->next_ptr;
      line_desc_ptr->next_ptr->prev_ptr = lower_ptr;

      return (line_desc_ptr);
    }
  } /* if lower_ptr */

  if (upper_ptr != NULL)
  {
    line_desc_ptr = upper_ptr->prev_ptr;

    if (line_desc_ptr == line_desc_ptr->prev_ptr)
    {
      return (NULL);
    }
    else
    {
      upper_ptr->prev_ptr               = line_desc_ptr->prev_ptr;
      line_desc_ptr->prev_ptr->next_ptr = upper_ptr;

      return (line_desc_ptr);
    }
  } /* if upper_ptr */

  return (NULL);
} /* get_line_desc */

/****************************************************************************/

FUNCTION LOCAL T_LINES_DESC *get_free_desc
  (
    IN    VOID
  )
  /*------------------------------------------------------------------------*/
  /* FUNCTIONAL_DESCRIPTION                                                 */
  /*------------------------------------------------------------------------*/
  /* - delivers a blank INI-file description element or NULL when out of    */
  /*   memory                                                               */
  /*------------------------------------------------------------------------*/
{
  LOCAL_VARIABLES

  register T_LINES_DESC  *line_desc_ptr;


  FUNCTION_BODY

  if ((line_desc_ptr = get_line_desc (&free_lines_head, NULL)) == NULL)
  {
    allocate_memory_block ();
    line_desc_ptr = get_line_desc (&free_lines_head, NULL);
  } /* if */

  return (line_desc_ptr);
} /* get_free_desc */

/****************************************************************************/

FUNCTION LOCAL VOID put_free_desc
  (
    IN  T_LINES_DESC  *line_desc_ptr
  )
  /*------------------------------------------------------------------------*/
  /* FUNCTIONAL_DESCRIPTION                                                 */
  /*------------------------------------------------------------------------*/
  /* - frees the via line_desc_ptr specified INI-file description entry     */
  /* - it will be added to the free_lines elements chain                    */
  /*------------------------------------------------------------------------*/
{
  LOCAL_VARIABLES

  FUNCTION_BODY

  line_desc_ptr->prev_ptr->next_ptr = line_desc_ptr->next_ptr;
  line_desc_ptr->next_ptr->prev_ptr = line_desc_ptr->prev_ptr;

  line_desc_ptr->line [0]           = EOS;
  line_desc_ptr->line_length        = 0;
  line_desc_ptr->line_type          = PROFILE_EMPTY;

  put_line_desc (line_desc_ptr, &free_lines_head, free_lines_head.next_ptr);
} /* put_free_desc */


/****************************************************************************/


FUNCTION LOCAL BOOL allocate_memory_block
  (
    IN VOID
  )
  /*------------------------------------------------------------------------*/
  /* FUNCTIONAL_DESCRIPTION                                                 */
  /*------------------------------------------------------------------------*/
  /* - creates new INI-file description elements and puts them to the       */
  /*   free_lines description chain                                         */
  /*------------------------------------------------------------------------*/
{
  LOCAL_VARIABLES

  register T_MEMORY_DESC   *memory_desc_ptr;
  VOID                     *memory_ptr;

  register T_LINES_DESC    *lines_desc_ptr;

  register int             i;


  FUNCTION_BODY

  memory_ptr = (VOID    *) malloc (sizeof (T_LINES_DESC) * MAX_MEMORY_LINES + sizeof (T_MEMORY_DESC));

  if (memory_ptr != NULL)
  {
    memory_desc_ptr = &memory_desc_head;

    while (memory_desc_ptr->next_ptr != NULL)
    {
      memory_desc_ptr = memory_desc_ptr->next_ptr;
    }

    memory_desc_ptr->next_ptr   = (T_MEMORY_DESC*) memory_ptr;
    memory_desc_ptr             = (T_MEMORY_DESC*) memory_ptr;
    memory_desc_ptr->next_ptr   = NULL;
    memory_desc_ptr->memory_ptr = memory_ptr;

    lines_desc_ptr              = (T_LINES_DESC*) ((char*) memory_ptr + sizeof (T_MEMORY_DESC));

    for (i = 0; i < MAX_MEMORY_LINES; i++)
    {
      lines_desc_ptr->line [0]    = EOS;
      lines_desc_ptr->line_length = 0;
      lines_desc_ptr->line_type   = PROFILE_EMPTY;
      put_line_desc (lines_desc_ptr, &free_lines_head, free_lines_head.next_ptr);
      lines_desc_ptr++;
    } /* for */

    return (TRUE);
  } /* if memory_ptr */

  return (FALSE);
} /* allocate_memory_block */

/****************************************************************************/

FUNCTION LOCAL VOID free_memory_blocks
  (
    IN    T_MEMORY_DESC  *memory_desc_ptr
  )
  /*------------------------------------------------------------------------*/
  /* FUNCTIONAL_DESCRIPTION                                                 */
  /*------------------------------------------------------------------------*/
  /* - frees the via allocate_memory_blocks created INI-file description    */
  /*   blocks                                                               */
  /*------------------------------------------------------------------------*/
{
  LOCAL_VARIABLES

  VOID *memory_ptr;


  FUNCTION_BODY

  while (memory_desc_ptr != NULL)
  {
    memory_ptr      = memory_desc_ptr->memory_ptr;
    memory_desc_ptr = memory_desc_ptr->next_ptr;

    free (memory_ptr);
  }
} /* free_memory_blocks */



/****************************************************************************/
/*** READ *******************************************************************/
/****************************************************************************/

FUNCTION GLOBAL BOOL profile_read
  (
    IN  const char   *profile_path
  )
  /*------------------------------------------------------------------------*/
  /* FUNCTIONAL_DESCRIPTION                                                 */
  /*------------------------------------------------------------------------*/
  /* - reads the via profile_path specified INI-file                        */
  /* - evaluates the content                                                */
  /*------------------------------------------------------------------------*/
{
  LOCAL_VARIABLES

  FILE            *file_ptr;

  int             ch;
  char            *c_ptr, c;
  size_t          length, lines;

  T_LINES_DESC    *line_desc_ptr;


  FUNCTION_BODY

  if ((file_ptr = fopen (profile_path, "r")) != NULL)
  {
    if ((line_desc_ptr = get_free_desc ()         ) == NULL) return (FALSE);

    c_ptr    = line_desc_ptr->line;
    length   = 0;
    lines    = 0;

    while ((ch = fgetc (file_ptr)) != EOF)
    {
      c = (char) ch;
      switch (c)
      {
        case LF :
        {
          if (length)
          {
            *c_ptr                     = EOS;
            line_desc_ptr->line_length = length;

            if (check_ini_line (line_desc_ptr))
            {
              put_line_desc (line_desc_ptr, used_lines_tail.prev_ptr, &used_lines_tail);
              lines++;

              if ((line_desc_ptr = get_free_desc ()) == NULL) c = EOF;
            }

            c_ptr    = line_desc_ptr->line;
            length   = 0;
          } /* if */

          break;
        }

        default :
        {
          if (length < MAX_PROFILE_WIDTH)
          {
            *c_ptr++ = c;
            length++;
          }
        }
      } /* switch c */
    } /* while !EOF */

    if (line_desc_ptr != NULL) put_free_desc (line_desc_ptr);

    fclose (file_ptr);
    return (TRUE);

  } /* if file_ptr */
  else
  {
    return(FALSE);
  }
} /* profile_read */


/****************************************************************************/
/*** INI MANIPULATION FUNCTIONS *********************************************/
/****************************************************************************/

FUNCTION GLOBAL size_t profile_get_string
  (
    IN    const char   *group_string_ptr,
    IN    const char   *name_string_ptr,
    INOUT char         *string_ptr
  )
  /*------------------------------------------------------------------------*/
  /* FUNCTIONAL_DESCRIPTION                                                 */
  /*------------------------------------------------------------------------*/
  /* - reads the via name_string_ptr specified INI-file entry if it is a    */
  /*   member of the via group_string_ptr selected section                  */
  /* - returns the length of the expected string or zero if the command     */
  /*   fails                                                                */
  /*------------------------------------------------------------------------*/
{
  LOCAL_VARIABLES

  register T_LINES_DESC *line_desc_ptr;

  register char         *str_ptr;
  register char         c;
  size_t                length;

  FUNCTION_BODY

  if ((line_desc_ptr = find_profile_group (group_string_ptr)) != NULL)
  {
    if ((line_desc_ptr = find_profile_name (line_desc_ptr, name_string_ptr)) != NULL)
    {
      str_ptr = line_desc_ptr->line + strlen (name_string_ptr);
      length  = 0;

      if (*str_ptr++ == '=')
      {
        do
        {
          c             = *str_ptr++;
          *string_ptr++ = c;

          length++;

        } while (c != EOS);

        return (length);
      } /* if profile string */
    } /* if name */
  } /* if group */

  return (0);
} /* profile_get_string */


/****************************************************************************/

FUNCTION GLOBAL int profile_get_int
  (
    IN    const char   *group_string_ptr,
    IN    const char   *name_string_ptr,
    INOUT int          default_value
  )
  /*------------------------------------------------------------------------*/
  /* FUNCTIONAL_DESCRIPTION                                                 */
  /*------------------------------------------------------------------------*/
  /* - reads the via name_string_ptr specified INI-file entry if it is a    */
  /*   member of via group_string_ptr selected section                      */
  /* - returns the value found in the INI-file or default_value             */
  /*------------------------------------------------------------------------*/
{
  LOCAL_VARIABLES

  char    str [MAX_STRING_LEN];
  int     value;


  FUNCTION_BODY

  if (profile_get_string (group_string_ptr, name_string_ptr, str))
  {
    if ((sscanf (str, "0x%x", &value)) == 1) return (value);
    if ((sscanf (str, "%d", &value)) == 1) return (value);
  }

  return (default_value);
} /* profile_get_int */


/*****************************************************************************/

FUNCTION GLOBAL VOID profile_open
  (
    IN    VOID
  )
  /*------------------------------------------------------------------------*/
  /* FUNCTIONAL_DESCRIPTION                                                 */
  /*------------------------------------------------------------------------*/
  /* - initializes the INI-file memory handling                             */
  /*------------------------------------------------------------------------*/
{
  LOCAL_VARIABLES

  FUNCTION_BODY

  memory_desc_head.next_ptr   = NULL;
  memory_desc_head.memory_ptr = NULL;

  used_lines_head.next_ptr    = &used_lines_tail;
  used_lines_head.prev_ptr    = &used_lines_head;
  strcpy (used_lines_head.line, "used_lines_head");
  used_lines_head.line_length = strlen ("used_lines_head");

  used_lines_tail.next_ptr    = &used_lines_tail;
  used_lines_tail.prev_ptr    = &used_lines_head;
  strcpy (used_lines_tail.line, "used_lines_tail");
  used_lines_tail.line_length = strlen ("used_lines_tail");

  free_lines_head.next_ptr    = &free_lines_tail;
  free_lines_head.prev_ptr    = &free_lines_head;
  strcpy (free_lines_head.line, "free_lines_head");
  free_lines_head.line_length = strlen ("free_lines_head");

  free_lines_tail.next_ptr    = &free_lines_tail;
  free_lines_tail.prev_ptr    = &free_lines_head;
  strcpy (free_lines_tail.line, "free_lines_tail");
  free_lines_tail.line_length = strlen ("free_lines_tail");

} /* open_profile */

/*****************************************************************************/

FUNCTION GLOBAL VOID profile_close
  (
    IN    VOID
  )
  /*------------------------------------------------------------------------*/
  /* FUNCTIONAL_DESCRIPTION                                                 */
  /*------------------------------------------------------------------------*/
  /* - terminates the INI-file handling                                     */
  /*------------------------------------------------------------------------*/
{
  LOCAL_VARIABLES

  FUNCTION_BODY

  free_memory_blocks (memory_desc_head.next_ptr);

  memory_desc_head.next_ptr = NULL;

} /* profile_close */


