/******************************************************************************
*                                                                             *
*                      SOFTING Industrial Automation GmbH                     *
*                          Richard-Reitzner-Allee 6                           *
*                                D-85540 Haar                                 *
*                        Phone: (++49)-(0)89-45656-0                          *
*                        Fax:   (++49)-(0)89-45656-399                        *
*                                                                             *
*           Copyright (C) SOFTING Industrial Automation GmbH 1995-2010        *
*                              All Rights Reserved                            *
*                                                                             *
*******************************************************************************

FILE_NAME               DPSFrame.c

PROJECT_NAME            PROFIBUS DP-Slave

MODULE                  DPSDemo

COMPONENT_LIBRARY       --------

AUTHOR                  SOFTING

VERSION                 1.30 / 2.06

DATE                    30.07.2009

STATUS

FUNCTIONAL_MODULE_DESCRIPTION

DPS Demo application
------------------------------

The DPSDemo application shows you how to access the functionality of the PROFIBUS PAPI.
It uses the functions of the PROFIBUS PAPI to communicate with the DP slave firmware.

Program structure
-----------------

The following hierarchichal structure with six threads is used:


main (DPSFrame.C): - check the commandline, initialize hardware settings
|                  - starts the scheduler
|                  - waits for the end of one of the created tasks to quit the program
|
|-- dpsframe_init (DPSFrame.C): - initialize all modules, exit- and abnormal
|   |                             termination handler
    |                           - creates the keyboard polling task
    |
    |-- dpssched_init (DPSSched.C): - initialize the task-, event- and timer control
    |   |
    |   |-- system_timer_task (DPSSched.C): - receive the ticks from DOS timer
    |   |                                     interrupt (55 ms)
    |                                       - looks for timer expirations and
    |                                         will inform the respective tasks via
    |                                         timer event
    |
    |-- dpsprofi_init (DPSProfi.C): - initialize the hardware and the PAPI
    |   |                           - initialize simulation of input data
    |   |
    |   |-- papi_task (DPSProfi.C): - controls the PAPI service interface
    |   |                           - send request and responses ( DPS_Init_Slave,
    |   |                             DPS_Get_Status )
    |   |                           - handle all events ( DPS_Set_Prm, etc. )
    |   |
    |   |-- input_task (DPSProfi.C): - simulate the input data of the Slave-Application
    |   |                            - increments a counter asynchron, after receiving an
    |   |                              event from the DOS timer interrupt
    |   |                            - read synchron the value of the BIOS-Timer
    |   |
    |   |-- output_task (DPSProfi.C): - after reveiving the cyclic timer event from the
    |   |                               scheduler system timer, read this task the output
    |                                   data and displays it
    |
    |-- keyboard_task (DPSFrame.C): - handles the keyboard input
                                    - display the possible commands after keys h or space
                                    - send an event to papi_task after key g
                                    - terminates after keys q, x or ESC


RELATED_DOCUMENTS

============================================================================*/
#include "keywords.h"

INCLUDES

#include <windows.h>
#include <conio.h>
#include <dos.h>

#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>


#include "pb_type.h"
#include "pb_conf.h"
#include "pb_if.h"
#include "pb_dp.h"
#include "pb_dps.h"
#include "pb_err.h"

#include "DPSFrame.h"
#include "DPSSched.h"
#include "DPSProfi.h"


/****************************************************************************/

LOCAL_DEFINES


/****************************************************************************/

LOCAL_TYPEDEFS

/****************************************************************************/

FUNCTION_DECLARATIONS

  LOCAL  PB_BOOL   check_commandline_parameters( int, char**, USIGN8*, USIGN8* );
  LOCAL  VOID      show_help_text              ( VOID );
  LOCAL  PB_BOOL   dpsframe_init               ( USIGN8, USIGN8 );
  LOCAL  VOID      dpsframe_exit               ( VOID );
  PUBLIC int       main                        ( int, char** );

/****************************************************************************/

EXPORT_DATA

/****************************************************************************/

IMPORT_DATA

/****************************************************************************/

LOCAL_DATA

  FILE FAR*                 h_log_file                        = NULL;

  /*--- GENERAL ADMINISTRATION MEMORY --------------------------------------*/

  static const char FAR*    STRING_LOG_FILENAME               = "DPSFrame.log";
  static const char FAR*    STRING_COPYRIGHT                  = "\n\tDPSDemo - Copyright (C) SOFTING Industrial Automation GmbH 1998-2010\n";

  static const USIGN16      MIN_DPRAM_ADDRESS                 = 0xC800;
  static const USIGN16      MAX_DPRAM_ADDRESS                 = 0xF400;
  static const USIGN16      MIN_IO_PORT_ADDRESS               = 0x0240;
  static const USIGN16      MAX_IO_PORT_ADDRESS               = 0x0344;

  static const USIGN8       SLAVE_ADD                         = 0;
  static const USIGN8       MIN_SLAVE_ADD                     = 0;

  static const int          MAX_NUMBER_COMMANDLINE_PARAMETERS = 3;

/****************************************************************************/
/*** AUXILLIARY FUNCTIONS ***************************************************/
/****************************************************************************/

FUNCTION LOCAL PB_BOOL check_commandline_parameters
  (
  IN  int                 argc,
  IN  char*               argv[],
  IN  USIGN8*             board_ptr,
  IN  USIGN8*             slave_add_ptr
  )
  /*------------------------------------------------------------------------*/
  /* FUNCTIONAL_DESCRIPTION                                                 */
  /*------------------------------------------------------------------------*/
  /* - This function checks the commandline for valid parameters.           */
  /* - The following 3 arguments are necessarily:                           */
  /*       -bBOARDNUMBERs   : 0..9           (Board-Number)                 */
  /*       -sSlave_Address  : 0..126, 255    (Address of this DP Station)   */
  /* - If the function detectes an invalid parameter or missing a parameter,*/
  /*   it will display an usage screen.                                     */
  /*------------------------------------------------------------------------*/
{
LOCAL_VARIABLES

  USIGN16      board;
  USIGN16      slave_add;

  int          arg_index;

  PB_BOOL      ret_val;

FUNCTION_BODY

  assert( argv != NULL );
  assert( board_ptr != NULL );
  assert( slave_add_ptr != NULL );

  ret_val = PB_FALSE;

  if( argc == MAX_NUMBER_COMMANDLINE_PARAMETERS )
  {
    ret_val = PB_TRUE;

    for( arg_index = 1; arg_index < argc; arg_index++ )
    {
      if( ( sscanf( argv[ arg_index ], "-b%hi", &board ) == 1 ) &&
          ( ret_val                                                  )    )
      {
        if( ( board >= 0 ) &&
            ( board <= 9 )    )
        {
          *board_ptr = (USIGN8) (board);
        }
        else
        {
          ret_val = PB_FALSE;
        }
      }
      else if( ( sscanf( argv[ arg_index ], "-s%hi", &slave_add ) == 1 ) &&
               ( ret_val                                                       )    )
      {
        if( ( ( slave_add >= MIN_SLAVE_ADD         ) &&
              ( slave_add <= DPS_DEFAULT_SLAVE_ADD )    ) ||
            ( slave_add == DPS_NON_VOLATILE_SLAVE_ADD   )    )
        {
          *slave_add_ptr = (USIGN8) slave_add;
        }
        else
        {
          ret_val = PB_FALSE;
        }
      }
      else
      {
        ret_val = PB_FALSE;
      }

    } /* for commandline parameters */
  } /* if commandline parameter */

  if( !ret_val )
  {
    show_help_text( );
  }

  return( ret_val );

} /* check_commandline_parameters */

/****************************************************************************/

FUNCTION LOCAL VOID show_help_text
  (
  IN  VOID
  )
  /*------------------------------------------------------------------------*/
  /* FUNCTIONAL_DESCRIPTION                                                 */
  /*------------------------------------------------------------------------*/
  /* - This usage screen will be displayed, if an invalid parameter is      */
  /*   detected or a parameter will be missed in the commandline.           */
  /*------------------------------------------------------------------------*/
{
LOCAL_VARIABLES

FUNCTION_BODY

  OUTPUT_STRING( "*****************************************************************************\n" );
  OUTPUT_STRING( "  PROFIBUS DPSDemo V 1.30/2.06 (C) 1998-2010 Softing Industrial Automation GmbH");
  OUTPUT_STRING( "                     Softing Industrial Automation GmbH");
  OUTPUT_STRING( "                          Richard-Reitzner-Allee 6");
  OUTPUT_STRING( "                             D-85540 Haar");
  OUTPUT_STRING( "               Tel: (++49) (0) 89 / 45 65 6 - 0");
  OUTPUT_STRING( "               Fax: (++49) (0) 89 / 45 65 6 - 399");
  OUTPUT_STRING( "                          http://www.softing.com\n");
  OUTPUT_STRING( "Usage  : DPSDemo -bBOARD_Number -sSlave_Address\n");
  OUTPUT_STRING( "BOARD_Number   : 0..9           (Board-Number)" );
  OUTPUT_STRING( "Slave_Address  : 0..126, 255    (Address of this DP Station)\n" );
  OUTPUT_STRING( "*****************************************************************************" );

} /* show_help_text */

/****************************************************************************/

FUNCTION LOCAL VOID show_commands
  (
  IN  VOID
  )
  /*------------------------------------------------------------------------*/
  /* FUNCTIONAL_DESCRIPTION                                                 */
  /*------------------------------------------------------------------------*/
  /* - This function displays the possible command to control the DP Slave  */
  /*   application (send request, end the application).                     */
  /*------------------------------------------------------------------------*/
{
LOCAL_VARIABLES

  char   output[ MAX_ERROR_DESCR_LEN ];

FUNCTION_BODY

  sprintf( output, "\n[ESC] [Q] [X]  : quit program" );
  OUTPUT_STRING( output );

  sprintf( output, "[G]            : DPS_Get_Status.req" );
  OUTPUT_STRING( output );

  sprintf( output, "[SPACE] [H]    : this help screen\n" );
  OUTPUT_STRING( output );

  return;

} /* show_commands */

/****************************************************************************/

FUNCTION GLOBAL VOID output_string
  (
    IN const char FAR* output
  )
  /*------------------------------------------------------------------------*/
  /* FUNCTIONAL_DESCRIPTION                                                 */
  /*------------------------------------------------------------------------*/
  /* - This function print a formatted String on Screen and into a LogFile, */
  /*   if a Handle for a File is available.                                 */
  /*------------------------------------------------------------------------*/
{
LOCAL_VARIABLES

  size_t      output_length;

FUNCTION_BODY

  assert( output != NULL );

  printf( "%s \n", output );

  if( h_log_file != NULL )
  {
    output_length = strlen( output );
    output_length = fwrite( output, sizeof( char ), output_length, h_log_file );
    output_length = fwrite( "\n", sizeof( char ), 1, h_log_file );

  } /* if handle for log_file available */

} /* output_string */

/****************************************************************************/
/*** CALLBACK FUNCTIONS *****************************************************/
/****************************************************************************/

FUNCTION LOCAL VOID keyboard_task
  (
    IN T_DPSSCHED_EVENT_ID    event_id
  )
  /*------------------------------------------------------------------------*/
  /* FUNCTIONAL_DESCRIPTION                                                 */
  /*------------------------------------------------------------------------*/
  /* - This Callback Function controls the keyboard input from user.        */
  /* - After key h or space a help screen will be displayed with all        */
  /*   possible commands.                                                   */
  /* - After the key g an event will be send to papi_task.                  */
  /* - The keys q, x or ESC stop the scheduler and terminate the            */
  /*   application.                                                         */
  /*------------------------------------------------------------------------*/
{
LOCAL_VARIABLES

  int   key;


FUNCTION_BODY

  assert (event_id == DPSSCHED_SCHEDULE_EVENT);

  if (_kbhit ())
  {
    key = _getch ();

    if (( key == 0 ) || ( key == 0xE0 ))
    {
      key = _getch ();
    }

    key = tolower (key);

    switch (key)
    {
      case 0x1B:
      case 'x' :
      case 'q' : dpssched_stop_schedule (); break;
      case 'g' : dpssched_send_event (DPSSCHED_SERVICE_IF_TASK_ID, DPSSCHED_GET_STATUS_EVENT); break;
      case 'h' :
      case ' ' : show_commands (); break;

    } /* switch key */
  } /* if kbhit */

  return;

} /* keyboard_task */

/****************************************************************************/
/*** INITIALIZATION / TERMINATION  DPS FRAME ********************************/
/****************************************************************************/

/*** cast conversion from 'const int' to 'void(__far __cdecl *)(int)' *******/

#pragma warning( disable: 4306 )
FUNCTION LOCAL PB_BOOL dpsframe_init
  (
  IN    USIGN8          board,
  IN    USIGN8          slave_add
  )
  /*------------------------------------------------------------------------*/
  /* FUNCTIONAL_DESCRIPTION                                                 */
  /*------------------------------------------------------------------------*/
  /* - This function is called to initialize all sub modules and install the*/
  /*   handler for abnormal termination (Control-C-Handler, Assertion-      */
  /*   Handler).                                                            */
  /* - A Logfile for the output of the application is opened.               */
  /* - At end  a task for keyboard control will be created.                 */
  /*------------------------------------------------------------------------*/
{
LOCAL_VARIABLES

  char    err_out[ MAX_ERROR_DESCR_LEN ];


FUNCTION_BODY

  OUTPUT_STRING( STRING_COPYRIGHT );
  OUTPUT_STRING( "initalize DPS Frame" );

  /*--- INITIALIZE LOCALS OF THIS MODULE -----------------------------------*/

  /*--- GET FILEHANDLE FOR LOGFILE -----------------------------------------*/

  h_log_file = fopen( STRING_LOG_FILENAME, "w" );

  if(h_log_file == NULL )
  {
    sprintf (err_out, "Cannot open LogFile %s in current working directory", STRING_LOG_FILENAME);
    OUTPUT_STRING( err_out );
  }


  /*--- INITIALIZE MODULE DPSSCHED -----------------------------------------*/

  if(! dpssched_init( ) )
  {
    OUTPUT_STRING( "Couldn't initalize module DPSSched" );
    return( PB_FALSE );
  }

  /*--- INITIALIZE MODULE DPSPROFI -----------------------------------------*/

  if(! dpsprofi_init( board, slave_add ) )
  {
    OUTPUT_STRING( "Couldn't initalize module DPSProfi" );
    return( PB_FALSE );
  }

  /*--- CREATE NEW TASK FOR KEYBOARD POLLING -------------------------------*/

  dpssched_create_task( DPSSCHED_KEYBOARD_TASK_ID, DPSSCHED_PRIORITY_1, keyboard_task );

  return( PB_TRUE );

} /* dpsframe_init */
#pragma warning( default: 4306 )

/****************************************************************************/

FUNCTION LOCAL VOID dpsframe_exit
  (
  IN  VOID
  )
  /*------------------------------------------------------------------------*/
  /* FUNCTIONAL_DESCRIPTION                                                 */
  /*------------------------------------------------------------------------*/
  /* - This function will terminate all sub modules and close the Logfile.  */
  /* - An abnormal program termination will use this function also. In this */
  /*   case do not:                                                         */
  /*       issue low-level or 'stdio.h' i/o-routines (e.g. printf, fread)   */
  /*       call heap routines or any routine use heap routines (e.g. malloc)*/
  /*       use any function that generates a system call (e.g. time)        */
  /*------------------------------------------------------------------------*/
{
LOCAL_VARIABLES

FUNCTION_BODY

  /*--- TERMINATE MODULE DPSPROFI AND DPSSCHED -----------------------------*/

  dpsprofi_exit( );
  dpssched_exit( );

  /*--- CLOSE THE LOG-FILE -------------------------------------------------*/

  if( h_log_file != NULL )
  {
    (VOID) fclose( h_log_file );
  }

  return;

} /* dpsframe_exit */



/****************************************************************************/
/*** MAIN FUNCTION **********************************************************/
/****************************************************************************/

FUNCTION PUBLIC int main
  (
    IN    int       argc,
    IN    char*     argv[]
  )
  /*------------------------------------------------------------------------*/
  /* FUNCTIONAL_DESCRIPTION                                                 */
  /*------------------------------------------------------------------------*/
  /* The main function controls the initialization and termination of the   */
  /* program.                                                               */
  /*------------------------------------------------------------------------*/
{
LOCAL_VARIABLES

  USIGN8          board;
  USIGN8          slave_add;

  int             result;


FUNCTION_BODY

  result = 0;

  if( check_commandline_parameters( argc, argv, &board, &slave_add ) )
  {
    /*--- INITIALIZE GLOBALS AND SUB MODULES -------------------------------*/

    if (dpsframe_init( board, slave_add ))
    {
        dpssched_start_schedule( );
        exit (0);
    } /* if init successful */

    dpsframe_exit ();  /* call dpsframe_exit also when dpsframe_init failes */

  } /* if valid commandline parameters */

  return (result);

} /* main */

/**** END O DPSDemo.c *******************************************************/
