/******************************************************************************
*                                                                             *
*                      SOFTING Industrial Automation GmbH                     *
*                          Richard-Reitzner-Allee 6                           *
*                                D-85540 Haar                                 *
*                        Phone: (++49)-(0)89-45656-0                          *
*                        Fax:   (++49)-(0)89-45656-399                        *
*                                                                             *
*           Copyright (C) SOFTING Industrial Automation GmbH 1995-2010        *
*                              All Rights Reserved                            *
*                                                                             *
*******************************************************************************

FILE_NAME               DPSProfi.c

PROJECT_NAME            PROFIBUS DP-Slave

MODULE                  DPSDemo

COMPONENT_LIBRARY       --------

AUTHOR                  SOFTING

VERSION                 1.30 / 2.06

DATE                    30.07.2009

STATUS

FUNCTIONAL_MODULE_DESCRIPTION

DPS Demo application
------------------------------

The DPSDemo application shows you how to access the functionality of the PROFIBUS PAPI.
It uses the functions of the PROFIBUS PAPI to communicate with the DP slave firmware.


Program structure
-----------------

The following hierarchichal structure with six threads is used:


main (DPSFrame.C): - check the commandline, initialize hardware settings
|                  - starts the scheduler
|                  - waits for the end of one of the created tasks to quit the program
|
|-- dpsframe_init (DPSFrame.C): - initialize all modules, exit- and abnormal
|   |                             termination handler
    |                           - creates the keyboard polling task
    |
    |-- dpssched_init (DPSSched.C): - initialize the task-, event- and timer control
    |   |
    |   |-- system_timer_task (DPSSched.C): - receive the ticks from DOS timer
    |   |                                     interrupt (55 ms)
    |                                       - looks for timer expirations and
    |                                         will inform the respective tasks via
    |                                         timer event
    |
    |-- dpsprofi_init (DPSProfi.C): - initialize the hardware and the PAPI
    |   |                           - initialize simulation of input data
    |   |
    |   |-- papi_task (DPSProfi.C): - controls the PAPI service interface
    |   |                           - send request and responses ( DPS_Init_Slave,
    |   |                             DPS_Get_Status )
    |   |                           - handle all events ( DPS_Set_Prm, etc. )
    |   |
    |   |-- input_task (DPSProfi.C): - simulate the input data of the Slave-Application
    |   |                            - increments a counter asynchron, after receiving an
    |   |                              event from the DOS timer interrupt
    |   |                            - read synchron the value of the BIOS-Timer
    |   |
    |   |-- output_task (DPSProfi.C): - after reveiving the cyclic timer event from the
    |   |                               scheduler system timer, read this task the output
    |                                   data and displays it
    |
    |-- keyboard_task (DPSFrame.C): - handles the keyboard input
                                    - display the possible commands after keys h or space
                                    - send an event to papi_task after key g
                                    - terminates after keys q, x or ESC


RELATED_DOCUMENTS
=============================================================================*/

#include "keywords.h"

INCLUDES

#ifdef WIN32
 #include <windows.h>
 #include <conio.h>
 #include <dos.h>
#endif

#include <stdio.h>
#include <string.h>
#include <assert.h>

#include "pb_type.h"
#include "pb_conf.h"
#include "pb_if.h"
#include "pb_dp.h"
#include "pb_dps.h"
#include "pb_err.h"

#include "dpsframe.h"
#include "dpssched.h"
#include "dpsprofi.h"

/****************************************************************************/

LOCAL_DEFINES

  /*--- CFG IDENTIFIER DEFINITIONS (EN 50170/DP) ---------------------------*/

  #define DP_IDENTIFIER_CONSISTENCY   0x80
  #define DP_IDENTIFIER_LEN           0x40

  #define DP_IDENTIFIER_IO_SPECIAL    0x00
  #define DP_IDENTIFIER_IO_INP        0x10
  #define DP_IDENTIFIER_IO_OUTP       0x20
  #define DP_IDENTIFIER_IO_INP_OUTP   0x30
  #define DP_IDENTIFIER_IO            0x30

  #define DP_IDENTIFIER_DATA_LEN      0x0F
  #define DP_IDENTIFIER_SPEC_LEN      0x3F

  #define DP_IDENTIFIER_FREE_PLACE    0x00
  #define DP_IDENTIFIER_INP_LEN       0x40
  #define DP_IDENTIFIER_OUTP_LEN      0x80
  #define DP_IDENTIFIER_INP_OUTP_LEN  0xC0
  #define DP_IDENTIFIER_SPECIAL       0xC0

  /*--- AUXILIARY CFG CONVERTION MACROS ------------------------------------*/

  #define _DP_GET_IDENTIFIER_DATA_LEN(idf_ptr)                               \
          (                                                                  \
              (*idf_ptr & DP_IDENTIFIER_LEN)                ?                \
            (((*idf_ptr & DP_IDENTIFIER_DATA_LEN) + 1) * 2) :                \
             ((*idf_ptr & DP_IDENTIFIER_DATA_LEN) + 1)                       \
          )

  #define _DP_GET_IDENTIFIER_SPEC_LEN(idf_ptr)                               \
          (                                                                  \
              (*idf_ptr & DP_IDENTIFIER_LEN)                ?                \
            (((*idf_ptr & DP_IDENTIFIER_SPEC_LEN) + 1) * 2) :                \
             ((*idf_ptr & DP_IDENTIFIER_SPEC_LEN) + 1)                       \
          )

  #define _DP_GET_IDENTIFIER_USER_LEN(idf_ptr)                               \
          (                                                                  \
            ( ((*idf_ptr & DP_IDENTIFIER_DATA_LEN) == 0x00) ||               \
              ((*idf_ptr & DP_IDENTIFIER_DATA_LEN) == 0x0F) ) ?              \
               (0                                )            :              \
               (*idf_ptr & DP_IDENTIFIER_DATA_LEN)                           \
          )

/****************************************************************************/

LOCAL_TYPEDEFS

/****************************************************************************/

FUNCTION_DECLARATIONS

extern INT16 CALL_CONV profi_snd_idle (void);

/****************************************************************************/

EXPORT_DATA

/****************************************************************************/

IMPORT_DATA

/****************************************************************************/

LOCAL_DATA

  static const USIGN8       MIN_INTERFACE_ERR_LEN                    = 60;
  static const USIGN8       MIN_DPS_ERR_LEN                          = 90;
  static const USIGN8       MIN_LAYER_SERVICE_LEN                    = 80;
  static const USIGN8       MIN_EXCEPTION_LEN                        = 150;

  static const USIGN32      BIOS_TIME_COUNTER                        = 0x0000046CL;

  static const USIGN32      DPS_CYCLIC_OUTPUT_TIMER_DURATION         = 1000ul;

  /*--- DEFAULT VALUES FOR SERVICE DESCRIPTION BLOCK -----------------------*/

  static const USIGN8       DPS_DEFAULT_INVOKE_ID                    = 0;
  static const USIGN16      DPS_DEFAULT_COMM_REF                     = 0;

  /*--- DEFAULT VALUES FOR DPS_INIT_SLAVE ----------------------------------*/

  static const USIGN8       DPS_DEFAULT_INPUT_LEN                    = 180;
  static const USIGN8       DPS_DEFAULT_OUTPUT_LEN                   = 180;
  static const USIGN8       DPS_DEFAULT_CFG_DATA_LEN                 = 32;
  static const USIGN8       DPS_DEFAULT_USR_PRM_DATA_LEN             = 32;
  static const USIGN8       DPS_DEFAULT_EXT_DIAG_DATA_LEN            = 32;
  static const USIGN8       DPS_DEFAULT_ADDRESS_DATA_LEN             = 28;


  static const USIGN8       DPS_DEFAULT_CFG_DATA []                  =
                            {
                              0x10, 0xD1, 0x21
                            };  /* 1byte input, 2words input, 2bytes output */

  /*--- TIMER INTERRUPT SIMULATES MEASURE VALUE ----------------------------*/

  static UINT               dpsprofi_timer_id;

  /*--- DP SLAVE INPUT/OUTPUT DATA -----------------------------------------*/

  static USIGN8             input_data [DP_MAX_INPUT_DATA_LEN];
  static USIGN8             output_data [DP_MAX_OUTPUT_DATA_LEN];

  static USIGN8             input_length  = 0;

  /*--- AUXILIARY VARIABLES ------------------------------------------------*/

  static char               out_string[ MAX_ERROR_DESCR_LEN ];

  static USIGN32            actual_time = 0;

/****************************************************************************/
/*** AUXILLIARY FUNCTIONS ***************************************************/
/****************************************************************************/

FUNCTION LOCAL INT16 convert_cfg_to_length
  (
  IN    const USIGN8 FAR*   cfg_data_ptr,
  IN    USIGN8              cfg_data_len,
  OUT   USIGN8 FAR*         number_inputs_ptr,
  OUT   USIGN8 FAR*         number_outputs_ptr
  )
  /*------------------------------------------------------------------------*/
  /* FUNCTIONAL_DESCRIPTION                                                 */
  /*------------------------------------------------------------------------*/
  /* This function converts DP configuration data (CFG) into length         */
  /* information. The result is the number of input and output bytes.       */
  /* The coding definition of this function is taken from the DP standard   */
  /* (EN 50170 or DIN 19245-3) and described with service DDLM_Chk_Cfg.     */
  /*                                                                        */
  /* The bytes of the CFG data are also called 'identifier' or 'special     */
  /* identifier' in case of the second coding mechanism.                    */
  /* (German: 'Kennung', 'Kennungsbyte' oder 'Spezielles Kennungsformat')   */
  /*------------------------------------------------------------------------*/
{
LOCAL_VARIABLES

  int       cfg_index;
  int       size;
  int       length;

  int       number_inputs;
  int       number_outputs;

  INT16     ret_val;

FUNCTION_BODY

  /*--- check input data ---------------------------------------------------*/

  assert( cfg_data_ptr != NULL );
  assert( ( cfg_data_len >= DPS_MIN_CFG_DATA_LEN ) &&
          ( cfg_data_len <= DPS_MAX_CFG_DATA_LEN )    );

  /*--- convert CFG to number of bytes -------------------------------------*/

  number_inputs  = 0;
  number_outputs = 0;
  cfg_index      = 0;
  ret_val        = E_OK;

  while (cfg_index < cfg_data_len)
  {
    switch (cfg_data_ptr [0] & DP_IDENTIFIER_IO)
    {
      case DP_IDENTIFIER_IO_INP:
      {
        length         = _DP_GET_IDENTIFIER_DATA_LEN (cfg_data_ptr);
        number_inputs += length;
        cfg_data_ptr++;
        cfg_index++;
        break;
      }

      case DP_IDENTIFIER_IO_OUTP:
      {
        length          = _DP_GET_IDENTIFIER_DATA_LEN (cfg_data_ptr);
        number_outputs += length;
        cfg_data_ptr++;
        cfg_index++;
        break;
      }

      case DP_IDENTIFIER_IO_INP_OUTP:
      {
        length          = _DP_GET_IDENTIFIER_DATA_LEN (cfg_data_ptr);
        number_inputs  += length;
        length          = _DP_GET_IDENTIFIER_DATA_LEN (cfg_data_ptr);
        number_outputs += length;
        cfg_data_ptr++;
        cfg_index++;
        break;
      }

      case DP_IDENTIFIER_IO_SPECIAL:
      {
        switch (cfg_data_ptr [0] & DP_IDENTIFIER_SPECIAL)
        {
          case DP_IDENTIFIER_INP_LEN:
          {
            length         = _DP_GET_IDENTIFIER_SPEC_LEN (&cfg_data_ptr [1]);
            number_inputs += length;
            size          = 2 + _DP_GET_IDENTIFIER_USER_LEN (cfg_data_ptr);
            cfg_data_ptr  += size;
            cfg_index     += size;
            break;
          }

          case DP_IDENTIFIER_OUTP_LEN:
          {
            length         = _DP_GET_IDENTIFIER_SPEC_LEN (&cfg_data_ptr [1]);
            number_outputs += length;
            size           = 2 + _DP_GET_IDENTIFIER_USER_LEN (cfg_data_ptr);
            cfg_data_ptr   += size;
            cfg_index      += size;
            break;
          }

          case DP_IDENTIFIER_INP_OUTP_LEN:
          {
            length          = _DP_GET_IDENTIFIER_SPEC_LEN (&cfg_data_ptr [1]);
            number_outputs += length;
            length          = _DP_GET_IDENTIFIER_SPEC_LEN (&cfg_data_ptr [2]);
            number_inputs  += length;
            size            = 3 + _DP_GET_IDENTIFIER_USER_LEN (cfg_data_ptr);
            cfg_data_ptr   += size;
            cfg_index      += size;
            break;
          }

          case DP_IDENTIFIER_FREE_PLACE:
          {
            size          = 1 + _DP_GET_IDENTIFIER_USER_LEN (cfg_data_ptr);
            cfg_data_ptr += size;
            cfg_index    += size;
            break;
          }
        } /* switch DP_IDENTIFIER_SPECIAL */

        break;
      } /* case DP_IDENTIFIER_IO_SPECIAL */
    } /* switch DP_IDENTIFIER_IO */
  } /* while cfg_data */

  /*--- return temporary values --------------------------------------------*/

  if( number_inputs <= DPS_DEFAULT_INPUT_LEN )
  {
    *number_inputs_ptr  = (USIGN8) number_inputs;
  }
  else
  {
    *number_inputs_ptr = 0;
    ret_val            = E_DPS_IV | E_DPS_WRONG_CFG_LEN;
  }

  if( number_outputs <= DPS_DEFAULT_OUTPUT_LEN )
  {
    *number_outputs_ptr = (USIGN8) number_outputs;
  }
  else
  {
    *number_outputs_ptr = 0;
    ret_val             = E_DPS_IV | E_DPS_WRONG_CFG_LEN;
  }

  return( ret_val );

} /* convert_cfg_to_length */

/****************************************************************************/

FUNCTION LOCAL VOID show_hex_dump
  (
  IN  USIGN8 FAR*   dump_ptr,
  IN  USIGN8        length
  )
  /*------------------------------------------------------------------------*/
  /* FUNCTIONAL_DESCRIPTION                                                 */
  /*------------------------------------------------------------------------*/
  /* - This function display the data up to 16 bytes per line in hexadecimal*/
  /*   format. If the length of data is greater then 16 bytes, the data     */
  /*   will be splitted in blocks of 16 bytes and each block will be        */
  /*   displayed in a separate line.                                        */
  /*------------------------------------------------------------------------*/
{
LOCAL_VARIABLES

  int   i;
  int   line;

  char  line_string [256];
  char  number_string [16];


FUNCTION_BODY

  assert( dump_ptr != NULL );
  assert( length <= DP_MAX_TELEGRAM_LEN);

  line = 0;

  sprintf (line_string, "[%02d]: ", line);

  for (i = 0; i < length; i++)
  {
    sprintf (number_string, "0x%02hX ", dump_ptr [i]);
    strcat (line_string, number_string);

    if ((i % 16 == 0) && (i > 0))
    {
      OUTPUT_STRING (line_string);
      line++;
      sprintf (line_string, "[%02d]: ", line);
    }
  }

  OUTPUT_STRING (line_string);

  return;

} /* show_hex_dump */

/****************************************************************************/

FUNCTION GLOBAL VOID get_interface_error_string
  (
    IN    INT16     error_code,
    INOUT char FAR* err_string_ptr,
    IN    USIGN8    length
  )
  /*------------------------------------------------------------------------*/
  /* FUNCTIONAL_DESCRIPTION                                                 */
  /*------------------------------------------------------------------------*/
  /* - This function map the error code from papi interface to a textual    */
  /*   error description.                                                   */
  /*------------------------------------------------------------------------*/
{
LOCAL_VARIABLES

FUNCTION_BODY

  assert( err_string_ptr != NULL );
  assert( length >= MIN_INTERFACE_ERR_LEN );

  switch( error_code )
  {
    case E_OK                            : break;
    case E_IF_FATAL_ERROR                : strcpy( err_string_ptr, "\tFBAPI ERROR: unrecoverable error on board "          ); break;
    case E_IF_INIT_INVALID_PARAMETER     : strcpy( err_string_ptr, "\tFBAPI ERROR: invalid initialize parameter "          ); break;
    case E_IF_NO_CNTRL_RES               : strcpy( err_string_ptr, "\tFBAPI ERROR: controller does not respond "           ); break;
    case E_IF_INVALID_CNTRL_TYPE_VERSION : strcpy( err_string_ptr, "\tFBAPI ERROR: invalid controller type or SW version " ); break;
    case E_IF_INVALID_LAYER              : strcpy( err_string_ptr, "\tFBAPI ERROR: invalid layer "                         ); break;
    case E_IF_INVALID_SERVICE            : strcpy( err_string_ptr, "\tFBAPI ERROR: invalid service identifier "            ); break;
    case E_IF_INVALID_PRIMITIVE          : strcpy( err_string_ptr, "\tFBAPI ERROR: invalid service primitive "             ); break;
    case E_IF_INVALID_DATA_SIZE          : strcpy( err_string_ptr, "\tFBAPI ERROR: not enough cmi data block memory "      ); break;
    case E_IF_INVALID_COMM_REF           : strcpy( err_string_ptr, "\tFBAPI ERROR: invalid communication reference "       ); break;
    case E_IF_INVALID_CMI_CALL           : strcpy( err_string_ptr, "\tFBAPI ERROR: invalid CMI call "                      ); break;
    case E_IF_CMI_ERROR                  : strcpy( err_string_ptr, "\tFBAPI ERROR: error occured in CMI "                  ); break;
    case E_IF_RESOURCE_UNAVAILABLE       : strcpy( err_string_ptr, "\tFBAPI ERROR: no resource available "                 ); break;
    case E_IF_NO_PARALLEL_SERVICES       : strcpy( err_string_ptr, "\tFBAPI ERROR: no parallel services allowed "          ); break;
    case E_IF_SERVICE_CONSTR_CONFLICT    : strcpy( err_string_ptr, "\tFBAPI ERROR: service temporary not executable "      ); break;
    case E_IF_SERVICE_NOT_SUPPORTED      : strcpy( err_string_ptr, "\tFBAPI ERROR: service not supported "                 ); break;
    case E_IF_SERVICE_NOT_EXECUTABLE     : strcpy( err_string_ptr, "\tFBAPI ERROR: service not executable "                ); break;
    case E_IF_NO_CNTRL_PRESENT           : strcpy( err_string_ptr, "\tFBAPI ERROR: no controller present "                 ); break;
    case E_IF_INVALID_PARAMETER          : strcpy( err_string_ptr, "\tFBAPI ERROR: invalid parameter "                     ); break;
    case E_IF_INIT_FAILED                : strcpy( err_string_ptr, "\tFBAPI ERROR: init failed "                           ); break;
    case E_IF_EXIT_FAILED                : strcpy( err_string_ptr, "\tFBAPI ERROR: exit failed "                           ); break;
    case E_IF_PAPI_NOT_INITIALIZED       : strcpy( err_string_ptr, "\tFBAPI ERROR: PAPI not initialized "                  ); break;
    case E_IF_SOCKET_ERROR               : sprintf( err_string_ptr, "\tFBAPI ERROR: TCP socket error 0x%04X",WSAGetLastError() ); break;
    case E_IF_OS_ERROR                   : sprintf( err_string_ptr, "\tFBAPI ERROR: OS error 0x%04X",GetLastError() ); break;
    default                              : sprintf( err_string_ptr, "\tFBAPI ERROR: unknown error code 0x%04X", error_code  ); break;
  } /* end switch( error_code ) */

} /* get_interface_error_string */

/****************************************************************************/

FUNCTION GLOBAL VOID get_dps_error_string
  (
    IN    USIGN16   error_code,
    INOUT char FAR* err_string_ptr,
    IN    USIGN8    length
  )
  /*------------------------------------------------------------------------*/
  /* FUNCTIONAL_DESCRIPTION                                                 */
  /*------------------------------------------------------------------------*/
  /* - This function map the error code from dps protocol stack to a textual*/
  /*   error description.                                                   */
  /*------------------------------------------------------------------------*/
{
LOCAL_VARIABLES

  USIGN16 service_result;
  USIGN16 error_extention;

FUNCTION_BODY

  assert( err_string_ptr != NULL );
  assert( length >= MIN_DPS_ERR_LEN );

  service_result  = (USIGN16) ( error_code & 0x00FF );
  error_extention = (USIGN16) ( error_code & 0xFF00 );

  switch( service_result )
  {
    case E_DPS_OK: break;
    case E_DPS_UE: strcpy( err_string_ptr, "\tDPS ERROR: remote user error "               ); break;
    case E_DPS_RR: strcpy( err_string_ptr, "\tDPS ERROR: remote resource insufficient "    ); break;
    case E_DPS_RS: strcpy( err_string_ptr, "\tDPS ERROR: remote service/SAP deactivated "  ); break;
    case E_DPS_RA: strcpy( err_string_ptr, "\tDPS ERROR: access of remote SAP blocked "    ); break;
    case E_DPS_NA: strcpy( err_string_ptr, "\tDPS ERROR: no reaction from remote station " ); break;
    case E_DPS_DS: strcpy( err_string_ptr, "\tDPS ERROR: local entity disconnected "       ); break;
    case E_DPS_NO: strcpy( err_string_ptr, "\tDPS ERROR: not possible in this state "      ); break;
    case E_DPS_LR: strcpy( err_string_ptr, "\tDPS ERROR: local resource not available "    ); break;
    case E_DPS_IV: strcpy( err_string_ptr, "\tDPS ERROR: invalid parameter in request "    ); break;
    case E_DPS_TO: strcpy( err_string_ptr, "\tDPS ERROR: timeout expired "                 ); break;
    case E_DPS_FE: strcpy( err_string_ptr, "\tDPS ERROR: format error in request frame "   ); break;
    case E_DPS_NI: strcpy( err_string_ptr, "\tDPS ERROR: function not implemented "        ); break;
    case E_DPS_AD: strcpy( err_string_ptr, "\tDPS ERROR: access denied "                   ); break;
    case E_DPS_EA: strcpy( err_string_ptr, "\tDPS ERROR: area too large "                  ); break;
    case E_DPS_LE: strcpy( err_string_ptr, "\tDPS ERROR: data block length exceeded "      ); break;
    case E_DPS_RE: strcpy( err_string_ptr, "\tDPS ERROR: format error in response frame "  ); break;
    case E_DPS_IP: strcpy( err_string_ptr, "\tDPS ERROR: invalid parameter "               ); break;
    case E_DPS_SC: strcpy( err_string_ptr, "\tDPS ERROR: sequence conflict "               ); break;
    case E_DPS_SE: strcpy( err_string_ptr, "\tDPS ERROR: sequence error "                  ); break;
    case E_DPS_NE: strcpy( err_string_ptr, "\tDPS ERROR: area non-existent "               ); break;
    case E_DPS_DI: strcpy( err_string_ptr, "\tDPS ERROR: data incomplete "                 ); break;
    case E_DPS_NC: strcpy( err_string_ptr, "\tDPS ERROR: not connected "                   ); break;
    default      : sprintf( err_string_ptr, "\tDPS ERROR: unknown error code 0x%04X", service_result ); break;
  } /* switch( service_result ) */

  switch( error_extention )
  {
    case E_DPS_OK                     : break;
    case E_DPS_DUPLICATED_SERVICE     : strcat( err_string_ptr, "[DUPLICATED_SERVICE] "       ); break;
    case E_DPS_WRONG_SLAVE_ADDRESS    : strcat( err_string_ptr, "[WRONG_SLAVE_ADDRESS] "      ); break;
    case E_DPS_WRONG_CFG_LEN          : strcat( err_string_ptr, "[WRONG_CFG_LEN] "            ); break;
    case E_DPS_WRONG_PRM_LEN          : strcat( err_string_ptr, "[WRONG_PRM_LEN] "            ); break;
    case E_DPS_WRONG_DIAG_LEN         : strcat( err_string_ptr, "[WRONG_DIAG_LEN] "           ); break;
    case E_DPS_WRONG_SSA_LEN          : strcat( err_string_ptr, "[WRONG_SSA_LEN] "            ); break;
    case E_DPS_WRONG_ENHANCED_INIT_LEN: strcat( err_string_ptr, "[WRONG_ENHANCED_INIT_LEN] "  ); break;
    case E_DPS_ILLEGAL_CFG_DATA       : strcat( err_string_ptr, "[ILLEGAL_CFG_DATA] "         ); break;
    case E_DPS_NOT_ENOUGH_FRAME_MEMORY: strcat( err_string_ptr, "[NOT_ENOUGH_FRAME_MEMORY] "  ); break;
    case E_DPS_WRONG_INPUT_LEN        : strcat( err_string_ptr, "[WRONG_INPUT_DATA_LENGTH] "  ); break;
    case E_DPS_WRONG_OUTPUT_LEN       : strcat( err_string_ptr, "[WRONG_OUTPUT_DATA_LENGTH] " ); break;
    case E_DPS_INVALID_DIAG_STATE     : strcat( err_string_ptr, "[INVALID_DIAG_STATE] "       ); break;
    case E_DPS_INVALID_EXT_DIAG_DATA  : strcat( err_string_ptr, "[INVALID_EXT_DIAG_DATA] "    ); break;
    case E_DPS_SSA_REQUIRED           : strcat( err_string_ptr, "[SSA_REQUIRED] "             ); break;
    default                           : strcat( err_string_ptr, "[UNKNOWN_EXTENTION] "        ); break;
  } /* switch( error_extention ) */

} /* get_dps_error_string */

/****************************************************************************/

FUNCTION LOCAL VOID get_layer_service_string
  (
    IN    USIGN8     layer,
    IN    USIGN8     service,
    INOUT char FAR*  out_string_ptr,
    IN    USIGN8     length
  )
  /*------------------------------------------------------------------------*/
  /* FUNCTIONAL_DESCRIPTION                                                 */
  /*------------------------------------------------------------------------*/
  /* - This function shows the layer and the service which is decoded in the*/
  /*   .con/.ind telegram.                                                  */
  /*------------------------------------------------------------------------*/
{
LOCAL_VARIABLES

  char       layer_service_string[ MAX_ERROR_DESCR_LEN ];

FUNCTION_BODY

  assert( out_string_ptr != NULL );
  assert( length >= MIN_LAYER_SERVICE_LEN );

  switch( layer )
  {

    case USR     : sprintf( layer_service_string, "\tlayer     : USR "                        ); break;
    case DPS     : sprintf( layer_service_string, "\tlayer     : DPS "                        ); break;
    case DPS_USR : sprintf( layer_service_string, "\tlayer     : DPS "                        ); break;
    case PAPI    : sprintf( layer_service_string, "\tlayer     : DPS "                        ); break;
    default      : sprintf( layer_service_string, "\tlayer     : unknown layer 0x%02X", layer ); break;

  } /* switch( layer ) */

  strcpy( out_string_ptr, layer_service_string );

  switch( service )
  {
    case DPS_INIT_SLAVE    : sprintf( layer_service_string, "\tservice   : dps_init_slave"                  ); break;
    case DPS_EXIT_SLAVE    : sprintf( layer_service_string, "\tservice   : dps_exit_slave"                  ); break;
    case DPS_CHK_CFG       : sprintf( layer_service_string, "\tservice   : dps_chk_cfg"                     ); break;
    case DPS_SET_PRM       : sprintf( layer_service_string, "\tservice   : dps_set_prm"                     ); break;
    case DPS_SET_SLAVE_ADD : sprintf( layer_service_string, "\tservice   : dps_set_slave_address"           ); break;
    case DPS_SLAVE_DIAG    : sprintf( layer_service_string, "\tservice   : dps_slave_diag"                  ); break;
    case DPS_GET_STATUS    : sprintf( layer_service_string, "\tservice   : dps_get_status"                  ); break;
    default                : sprintf( layer_service_string, "\tservice   : unknown service 0x%02X", service ); break;

  } /* switch( service ) */

  strcat( out_string_ptr, layer_service_string );

  return;

} /* get_layer_service_string */

/****************************************************************************/

FUNCTION LOCAL VOID get_fatal_error_string
  (
    IN    T_EXCEPTION FAR* exception_ptr,
    IN    char FAR*        err_string_ptr,
    IN    USIGN8           length
  )
  /*------------------------------------------------------------------------*/
  /* FUNCTIONAL_DESCRIPTION                                                 */
  /*------------------------------------------------------------------------*/
  /* - This function shows the description of an unrecoverable error on     */
  /*   board.                                                               */
  /*------------------------------------------------------------------------*/
{
LOCAL_VARIABLES

  char       fatal_err_string[ MAX_ERROR_DESCR_LEN ];

FUNCTION_BODY

  assert( exception_ptr != NULL );
  assert( err_string_ptr != NULL );
  assert( length >= MIN_EXCEPTION_LEN );

  sprintf (fatal_err_string, "\ttask_id    : 0x%04hX, %hu\n", exception_ptr->task_id, exception_ptr->task_id);
  strcpy( err_string_ptr, fatal_err_string );

  sprintf (fatal_err_string, "\tparameter 1: 0x%04hX, %u\n", exception_ptr->par1, exception_ptr->par1);
  strcat( err_string_ptr, fatal_err_string );

  sprintf (fatal_err_string, "\tparameter 2: 0x%04hX, %u\n", exception_ptr->par2, exception_ptr->par2);
  strcat( err_string_ptr, fatal_err_string );

  sprintf (fatal_err_string, "\tparameter 3: 0x%04hX, %u", exception_ptr->par3, exception_ptr->par3);
  strcat( err_string_ptr, fatal_err_string );

} /* get_fatal_error_string */

/****************************************************************************/

FUNCTION LOCAL VOID show_cfg_data
  (
  IN  T_DPS_CHK_CFG_IND FAR*   chk_cfg_ind_ptr
  )
  /*------------------------------------------------------------------------*/
  /* FUNCTIONAL_DESCRIPTION                                                 */
  /*------------------------------------------------------------------------*/
  /* -This function displays the data received with dps_chk_cfg.ind from    */
  /*  DP master.                                                            */
  /* -If the cfg_await_response is set true, the DP slave firmware await a  */
  /*  positive or negative repsonse from user.                              */
  /* -The coding definition of the cfg_data is taken from the DP standard   */
  /*  (EN 50170 or DIN 19245-3) and described with service DDLM_Chk_Cfg.    */
  /*                                                                        */
  /* -The bytes of the CFG data are also called 'identifier' or 'special    */
  /*  identifier' in case of the second coding mechanism.                   */
  /*  (German: 'Kennung', 'Kennungsbyte' oder 'Spezielles Kennungsformat')  */
  /*------------------------------------------------------------------------*/
{
LOCAL_VARIABLES

FUNCTION_BODY

  assert( chk_cfg_ind_ptr != NULL );

  out_string [0]  = '\0';

  if( chk_cfg_ind_ptr->status != E_DPS_OK )
  {
    get_dps_error_string( chk_cfg_ind_ptr->status, out_string, sizeof( out_string ) );
    OUTPUT_STRING( out_string );
  }
  sprintf( out_string, "cfg_data_len       : %u", chk_cfg_ind_ptr->cfg_data_len );
  OUTPUT_STRING( out_string );

  if( chk_cfg_ind_ptr->cfg_await_response )
  {
    OUTPUT_STRING( "cfg_await_response : TRUE" );

  }
  else
  {
    OUTPUT_STRING( "cfg_await_response : FALSE" );

  }

  OUTPUT_STRING( "cfg_data           : " );
  show_hex_dump( chk_cfg_ind_ptr->cfg_data, chk_cfg_ind_ptr->cfg_data_len );

  return;

} /* show_cfg_data */

/****************************************************************************/

FUNCTION LOCAL VOID show_prm_data
  (
  IN  T_DPS_SET_PRM_IND FAR*   dps_set_prm_ind_ptr
  )
  /*------------------------------------------------------------------------*/
  /* FUNCTIONAL_DESCRIPTION                                                 */
  /*------------------------------------------------------------------------*/
  /* -This function displays the data received with dps_set_prm.ind from    */
  /*  DP master.                                                            */
  /* -If the prm_await_response is set true, the DP slave firmware await a  */
  /*  positive or negative repsonse from user.                              */
  /* -The parameter station_state consist of bus specific data and DP-Slave */
  /*  specific data.                                                        */
  /* -The parameters wd_fact_1 and wd_fact_2 represent factors for setting  */
  /*  the watchdog control. The watchdog control in a DP-Slave takes care   */
  /*  that, if the master fails, the outputs fall in the safe state after   */
  /*  the expiration of this time.                                          */
  /* -The parameter min_tsdr is the minimum waiting time for a DP-Slave     */
  /*  until it is allowed to send the response frames to the DP-Master.     */
  /* -The DP-Slave accepts only parameter frames if the transmitted         */
  /*  ident_number is identical with the own ident_number, which is         */
  /*  established by PNO.                                                   */
  /* -With the parameter group_ident it is possible to build groups for the */
  /*  function DDLM_Global_Control. Each bit represents a own group.        */
  /* -The usr_prm_data are freely allocatable for DP-Slave specific         */
  /*  parameters. The meaning and the ranges of values are defined          */
  /*  application specific.                                                 */
  /*  !!! NOTE !!! The first byte is'nt freely allocatable, because the     */
  /*               ASIC SPC3 use bit0..bit2 for special internal setting.   */
  /*               Therefore do not use this byte.                          */
  /*------------------------------------------------------------------------*/
{
LOCAL_VARIABLES

  PB_BOOL       first;


FUNCTION_BODY

  assert( dps_set_prm_ind_ptr != NULL );

  first           = PB_TRUE;
  out_string [0]  = '\0';

  if( dps_set_prm_ind_ptr->status != E_DPS_OK )
  {
    get_dps_error_string( dps_set_prm_ind_ptr->status, out_string, sizeof( out_string ) );
    OUTPUT_STRING( out_string );
  }

  sprintf( out_string, "user_prm_data_len  : %u", dps_set_prm_ind_ptr->user_prm_data_len );
  OUTPUT_STRING( out_string );

  if( dps_set_prm_ind_ptr->prm_await_response )
  {
    OUTPUT_STRING( "prm_await_response : TRUE" );
  }
  else
  {
    OUTPUT_STRING( "prm_await_response : FALSE" );
  }

  if( dps_set_prm_ind_ptr->station_status & DP_PRM_LOCK_REQ )
  {
    strcpy ( out_string, "station_status     : LOCK" );

    first = PB_FALSE;
  }

  if( dps_set_prm_ind_ptr->station_status & DP_PRM_UNLOCK_REQ )
  {
    if( first )
      strcpy( out_string, "station_status     : UNLOCK" );
    else
      strcat( out_string, " | UNLOCK" );

    first = PB_FALSE;

  }
  if( dps_set_prm_ind_ptr->station_status & DP_PRM_SYNC_REQ )
  {
    if( first )
      strcpy( out_string, "station_status     : SYNC" );
    else
      strcat( out_string, " | SYNC" );

    first = PB_FALSE;

  }
  if( dps_set_prm_ind_ptr->station_status & DP_PRM_FREEZE_REQ )
  {
    if( first )
      strcpy( out_string, "station_status     : FREEZE" );
    else
      strcat( out_string, " | FREEZE" );

    first = PB_FALSE;

  }
  if( dps_set_prm_ind_ptr->station_status & DP_PRM_WD_ON )
  {
    if( first )
      strcpy( out_string, "station_status     : WD_ON" );
    else
      strcat( out_string, " | WD_ON" );

    first = PB_FALSE;

  }
  if( first )
  {
    sprintf( out_string, "station_status     : 0x%02X", dps_set_prm_ind_ptr->station_status );
  }

  OUTPUT_STRING( out_string );

  sprintf( out_string, "wd_fact_1          : %u", dps_set_prm_ind_ptr->wd_fact_1 );
  OUTPUT_STRING( out_string );

  sprintf( out_string, "wd_fact_2          : %u", dps_set_prm_ind_ptr->wd_fact_1 );
  OUTPUT_STRING( out_string );

  sprintf( out_string, "min_tsdr           : %u", dps_set_prm_ind_ptr->min_tsdr );
  OUTPUT_STRING( out_string );

  sprintf( out_string, "ident_number       : 0x%04X", dps_set_prm_ind_ptr->ident_number );
  OUTPUT_STRING( out_string );

  sprintf( out_string, "group_ident        : %u", dps_set_prm_ind_ptr->group_ident );
  OUTPUT_STRING( out_string );

  if( dps_set_prm_ind_ptr->user_prm_data_len > 0 )
  {
    first = PB_TRUE;

    if( dps_set_prm_ind_ptr->user_prm_data[ 0 ] & DPS_SPC3_USR_PRM_DISABLE_STARTBIT )
    {
      sprintf( out_string, "spc3_usr_prm_data  : DISABLE_STARTBIT" );

      first = PB_FALSE;

    }
    if( dps_set_prm_ind_ptr->user_prm_data[ 0 ] & DPS_SPC3_USR_PRM_DISABLE_STOPBIT )
    {
      if( first )
        sprintf( out_string, "spc3_usr_prm_data  : DISABLE_STOPBIT" );
      else
        strcat( out_string, " | DISABLE_STOPTBIT" );

      first = PB_FALSE;

    }
    if( dps_set_prm_ind_ptr->user_prm_data[ 0 ] & DPS_SPC3_USR_PRM_WD_BASE_1MS )
    {
      if( first )
        sprintf( out_string, "spc3_usr_prm_data  : WD_BASE_1MS" );
      else
        strcat( out_string, " | WD_BASE_1MS" );

      first = PB_FALSE;

    }
    if( first )
    {
      sprintf( out_string, "spc3_usr_prm_data  : 0x%02X", dps_set_prm_ind_ptr->user_prm_data[ 0 ] );
    }

    OUTPUT_STRING( out_string );
    OUTPUT_STRING( "user_prm_data      : " );
    show_hex_dump( dps_set_prm_ind_ptr->user_prm_data, dps_set_prm_ind_ptr->user_prm_data_len );

  } /* if user_prm_data_len */

  return;

} /* show_prm_data */

/****************************************************************************/

FUNCTION LOCAL VOID show_ssa_data
  (
  IN  T_DPS_SET_SLAVE_ADD_IND FAR*   set_slave_add_ind_ptr
  )
  /*------------------------------------------------------------------------*/
  /* FUNCTIONAL_DESCRIPTION                                                 */
  /*------------------------------------------------------------------------*/
  /* -This function displays the data received with dps_set_slave_add.ind   */
  /*  from DP master.                                                       */
  /* -The parameter new_slave_add contains DP slave address, has be set by  */
  /*  the firmware.                                                         */
  /* -The DP-Slave accepts only parameter frames if the transmitted         */
  /*  ident_number is identical with the own ident_number, which is         */
  /*  established by PNO.                                                   */
  /* -The parameter no_add_chg specifies whether it is allowed to change    */
  /*  the DP-Slave address again at a later date.                           */
  /* -The rem_slave_data contains user specific data.                       */
  /*------------------------------------------------------------------------*/
{
LOCAL_VARIABLES

FUNCTION_BODY

  assert (set_slave_add_ind_ptr != NULL);

  out_string [0]  = '\0';

  if( set_slave_add_ind_ptr->status != E_DPS_OK )
  {
    get_dps_error_string( set_slave_add_ind_ptr->status, out_string, sizeof( out_string ) );
    OUTPUT_STRING( out_string );
  }

  sprintf( out_string, "rem_slave_data_len : %u", set_slave_add_ind_ptr->rem_slave_data_len );
  OUTPUT_STRING( out_string );

  sprintf( out_string, "new_slave_add      : %u", set_slave_add_ind_ptr->new_slave_add );
  OUTPUT_STRING( out_string );

  sprintf( out_string, "ident_number       : 0x%04X", set_slave_add_ind_ptr->ident_number );
  OUTPUT_STRING( out_string );

  strcpy( out_string, "no_add_change      : " );
  if (set_slave_add_ind_ptr->no_add_chg)
  {
    strcat( out_string, "TRUE" );
  }
  else
  {
    strcat( out_string, "FALSE" );
  }

  OUTPUT_STRING( out_string );

  if( set_slave_add_ind_ptr->rem_slave_data_len > 0 )
  {
    OUTPUT_STRING( out_string );
    OUTPUT_STRING( "rem_slave_data     : " );
    show_hex_dump( set_slave_add_ind_ptr->rem_slave_data, set_slave_add_ind_ptr->rem_slave_data_len );
  }

  return;

} /* show_ssa_data */

/****************************************************************************/

FUNCTION LOCAL VOID show_get_status_data
  (
  IN  T_DPS_GET_STATUS_CON_IND FAR*   get_status_con_ind_ptr
  )
  /*------------------------------------------------------------------------*/
  /* FUNCTIONAL_DESCRIPTION                                                 */
  /*------------------------------------------------------------------------*/
  /* -This function displays the data received local with dps_get_status.   */
  /*  con/ind from DP slave firmware.                                       */
  /* -The parameter slave_state contains the status of firmware, if no      */
  /*  traffic on fieldbus the status is idle else the status could be       */
  /*  wait_prm, wait_cfg or data_exchange, if a master sends data to the    */
  /*  DP slave. If an internal error occur the status will be controller    */
  /*  error.                                                                */
  /* -The parameter diag_state indicates, if exists a diagnostic entry in   */
  /*  slave diag buffer. If the stat_diag bit is set the DP master must get */
  /*  data until the bit is set to false. If the ext_diag bit is set it     */
  /*  indicates an entry in slavespecific diag buffer. The ext_diag_overflow*/
  /*  bit is set if the size of diagnostic data is greater as the send      */
  /*  buffer.                                                               */
  /* -The DP-Slave accepts only parameter frames if the transmitted         */
  /*  ident_number is identical with the own ident_number, which is         */
  /*  established by PNO.                                                   */
  /* -The number of inputs and outputs are the result of the cfg_data.      */
  /* -The slave_add is the current address of the station below this station*/
  /*  sends and receives telegrams. The non_volatile_slave_add is the       */
  /*  address which is stored in non volatile memory, like eeprom or flash  */
  /*  and is set by the service ddlm_set_slave_add. from DP master.         */
  /* -The parameter master_add indicates the active station, which sends to */
  /*  and receives data from this one.                                      */
  /* -The baud_rate indicates the speed of filedbus traffic.                */
  /* -The master_mode show which commands the DP slave received with the    */
  /*  last global_control command.                                          */
  /* -prm_await_response indicate whether the user must acknowledge the     */
  /*  set_prm.ind.                                                          */
  /* -cfg_await_response indicate whether the user must acknowledge the     */
  /*  chk_cfg.ind.                                                          */
  /* -await_startup_inputs indicate whether the user must send values for   */
  /*  input data befor the DP slave can change state to data_exchange       */
  /*------------------------------------------------------------------------*/
{
LOCAL_VARIABLES

  PB_BOOL       first;

FUNCTION_BODY

  assert( get_status_con_ind_ptr != NULL );

  first           = PB_TRUE;
  out_string [0]  = '\0';

  if( get_status_con_ind_ptr->status != E_DPS_OK )
  {
    get_dps_error_string( get_status_con_ind_ptr->status, out_string, sizeof( out_string ) );
    OUTPUT_STRING( out_string );
  }

  strcpy( out_string, "slave_state            : " );
  switch (get_status_con_ind_ptr->slave_state)
  {
    case DPS_STATE_IDLE            : strcat( out_string, "IDLE"             ); break;
    case DPS_STATE_WAIT_PRM        : strcat( out_string, "WAIT_PRM"         ); break;
    case DPS_STATE_WAIT_CFG        : strcat( out_string, "WAIT_CFG"         ); break;
    case DPS_STATE_DATA_EXCHANGE   : strcat( out_string, "DATA_EXCHANGE"    ); break;
    case DPS_STATE_CONTROLLER_ERROR: strcat( out_string, "CONTROLLER_ERROR" ); break;
  }

  OUTPUT_STRING( out_string );

  strcpy( out_string, "diag_state             : " );
  if (get_status_con_ind_ptr->diag_state == 0)
  {
    strcat( out_string, "OK" );
  }
  if (get_status_con_ind_ptr->diag_state & DPS_DIAG_BIT_EXT_DIAG)
  {
    strcat( out_string, "EXT_DIAG" );

    first = PB_FALSE;
  }
  if (get_status_con_ind_ptr->diag_state & DPS_DIAG_BIT_STAT_DIAG)
  {
    if (!first)
    {
      strcat( out_string, " | " );
    }
    strcat( out_string, "STAT_DIAG" );

    first = PB_FALSE;
  }
  if (get_status_con_ind_ptr->diag_state & DPS_DIAG_BIT_EXT_DIAG_OVERFLOW)
  {
    if (!first)
    {
      strcat( out_string, " | " );
    }
    strcat( out_string, "EXT_DIAG_OVERFLOW" );

    first = PB_FALSE;
  }

  OUTPUT_STRING( out_string );

  sprintf( out_string, "ident_number           : 0x%04X", get_status_con_ind_ptr->ident_number );
  OUTPUT_STRING( out_string );

  sprintf( out_string, "number_inputs          : %u", get_status_con_ind_ptr->number_inputs );
  OUTPUT_STRING( out_string );

  sprintf( out_string, "number_outputs         : %u", get_status_con_ind_ptr->number_outputs );
  OUTPUT_STRING( out_string );

  sprintf( out_string, "slave_add              : %u", get_status_con_ind_ptr->slave_add );
  OUTPUT_STRING( out_string );

  sprintf( out_string, "non_volatile_slave_add : %u", get_status_con_ind_ptr->non_volatile_slave_add );
  OUTPUT_STRING( out_string );

  sprintf( out_string, "master_add             : %u", get_status_con_ind_ptr->master_add );
  OUTPUT_STRING( out_string );

  strcpy( out_string, "baud_rate              : " );
  switch (get_status_con_ind_ptr->baud_rate)
  {
    case DPS_KBAUD_9_6      : strcat( out_string, "9.6 kbits/s"            ); break;
    case DPS_KBAUD_19_2     : strcat( out_string, "19.2 kbits/s"           ); break;
    case DPS_KBAUD_93_75    : strcat( out_string, "93.75 kbits/s"          ); break;
    case DPS_KBAUD_187_5    : strcat( out_string, "187.5 kbits/s"          ); break;
    case DPS_KBAUD_500      : strcat( out_string, "500 kbits/s"            ); break;
    case DPS_MBAUD_1_5      : strcat( out_string, "1.5 Mbits/s"            ); break;
    case DPS_MBAUD_3        : strcat( out_string, "3 Mbits/s"              ); break;
    case DPS_MBAUD_6        : strcat( out_string, "6 Mbits/s"              ); break;
    case DPS_MBAUD_12       : strcat( out_string, "12 Mbits/s"             ); break;
    case DPS_KBAUD_31_25    : strcat( out_string, "31.25 kbits/s"          ); break;
    case DPS_KBAUD_45_45    : strcat( out_string, "45.45 kbits/s"          ); break;
    case DPS_NO_BUS_TRAFFIC : strcat( out_string, "NO TRAFFIC ON FIELDBUS" ); break;
    default                 : strcat( out_string, "UNKNOWN BAUDRATE"       ); break;
  } /* switch baudrate */

  OUTPUT_STRING( out_string );

  first           = PB_TRUE;
  strcpy( out_string, "master_mode            : " );
  if (get_status_con_ind_ptr->sync_enabled)
  {
    strcat( out_string, "SYNC" );

    first = PB_FALSE;
  }
  else
  {
    strcat( out_string, "UNSYNC" );

    first = PB_FALSE;
  }
  if (get_status_con_ind_ptr->freeze_enabled)
  {
    if (!first)
    {
      strcat( out_string, " | " );
    }
    strcat( out_string, "FREEZE" );

    first = PB_FALSE;
  }
  else
  {
    if (!first)
    {
      strcat( out_string, " | " );
    }
    strcat( out_string, "UNFREEZE" );

    first = PB_FALSE;
  }
  if (get_status_con_ind_ptr->clear_data)
  {
    if (!first)
    {
      strcat( out_string, " | " );
    }
    strcat( out_string, "CLEAR" );

    first = PB_FALSE;
  }
  else
  {
    if (!first)
    {
      strcat( out_string, " | " );
    }
    strcat( out_string, "OPERATE" );

    first = PB_FALSE;
  }

  OUTPUT_STRING( out_string );

  strcpy( out_string, "prm_await_response     : " );
  if (get_status_con_ind_ptr->prm_await_response)
  {
    strcat( out_string, "TRUE" );
  }
  else
  {
    strcat( out_string, "FALSE" );
  }

  OUTPUT_STRING( out_string );

  strcpy( out_string, "cfg_await_response     : " );
  if (get_status_con_ind_ptr->cfg_await_response)
  {
    strcat( out_string, "TRUE" );
  }
  else
  {
    strcat( out_string, "FALSE" );
  }

  OUTPUT_STRING( out_string );

  strcpy( out_string, "await_startup_inputs   : " );
  if (get_status_con_ind_ptr->await_startup_inputs)
  {
    strcat( out_string, "TRUE" );
  }
  else
  {
    strcat( out_string, "FALSE" );
  }

  OUTPUT_STRING( out_string );
  OUTPUT_STRING( "" );

  return;
}

/****************************************************************************/

FUNCTION LOCAL VOID show_service
  (
  IN  T_PROFI_SERVICE_DESCR*  slave_sdb,
  IN  T_DPSPROFI_CON_IND*     slave_con_ind
  )
  /*------------------------------------------------------------------------*/
  /* FUNCTIONAL_DESCRIPTION                                                 */
  /*------------------------------------------------------------------------*/
  /* -This function displays the services received by the PAPI function     */
  /*  profi_rcv_con_ind. It also shows the result of the servive, if a      */
  /*  confirmation was received.                                            */
  /*------------------------------------------------------------------------*/
{
LOCAL_VARIABLES

FUNCTION_BODY

  assert( slave_sdb != NULL );
  assert( slave_con_ind != NULL );

  switch( slave_sdb->service )
  {

    case DPS_INIT_SLAVE:
    {
      if( slave_sdb->result == POS )
      {
        OUTPUT_STRING( "dps_init_slave.con(+)" );
      }
      else
      {
        OUTPUT_STRING( "dps_init_slave.con(-)" );
        get_dps_error_string( slave_con_ind->dps_init_con.status, out_string, sizeof( out_string ) );
        OUTPUT_STRING( out_string );
      }

      sprintf( out_string, "remaining frame memory : %d", slave_con_ind->dps_init_con.remaining_frame_memory );
      OUTPUT_STRING( out_string );
      break;
    } /* case DPS_INIT_SLAVE */

    /*----------------------------------------------------------------------*/

    case DPS_EXIT_SLAVE:
    {
      if( slave_sdb->result == POS )
      {
        OUTPUT_STRING( "dps_exit_slave.con(+)" );
      }
      else
      {
        OUTPUT_STRING( "dps_exit_slave.con(-)" );
        get_dps_error_string( slave_con_ind->dps_exit_con.status, out_string, sizeof( out_string ) );
        OUTPUT_STRING( out_string );
      }
      break;
    }  /* case DPS_EXIT_SLAVE */

    /*----------------------------------------------------------------------*/

    case DPS_SLAVE_DIAG:
    {
      if( slave_sdb->result == POS )
      {
        OUTPUT_STRING( "dps_slave_diag.con(+)" );
      }
      else
      {
        OUTPUT_STRING( "dps_slave_diag.con(-)" );
        get_dps_error_string( slave_con_ind->dps_diag_con.status, out_string, sizeof( out_string ) );
        OUTPUT_STRING( out_string );
      }
      break;
    } /* case DPS_SLAVE_DIAG */

    /*----------------------------------------------------------------------*/

    case DPS_CHK_CFG:
    {
      if( slave_sdb->primitive == IND )
      {
         OUTPUT_STRING( "\ndps_chk_cfg.ind" );
         show_cfg_data( &slave_con_ind->dps_chk_cfg_ind );
      }
      else
      {
        OUTPUT_STRING( "invalid dps_chk_cfg.con/ind" );
      }
      break;
    }  /* case DPS_CHK_CFG */

    /*----------------------------------------------------------------------*/

    case DPS_SET_PRM:
    {
      if( slave_sdb->primitive == IND )
      {
        OUTPUT_STRING( "\ndps_set_prm.ind" );
        show_prm_data( &slave_con_ind->dps_set_prm_ind );
      }
      else
      {
        OUTPUT_STRING( "invalid dps_set_prm.con/ind" );
      }
      break;
    } /* case DPS_SET_PRM */

    /*----------------------------------------------------------------------*/

    case DPS_SET_SLAVE_ADD:
    {
      if( slave_sdb->primitive == IND )
      {
        OUTPUT_STRING( "\ndps_set_slave_add.ind" );
        show_ssa_data( &slave_con_ind->dps_set_slave_add_ind );
      }
      else
      {
        OUTPUT_STRING( "invalid dps_set_slave_add.con/ind" );
      }
      break;
      }
    /*----------------------------------------------------------------------*/

    case DPS_GET_STATUS:
    {
      if( ( slave_sdb->primitive == CON ) &&
          ( slave_sdb->result    == NEG )    )
      {
        OUTPUT_STRING( "dps_get_status.con(-)" );
        get_dps_error_string( slave_con_ind->dps_get_status_con_ind.status,
                         out_string,
                         sizeof( out_string ) );
        OUTPUT_STRING( out_string );
      }
      else
      {
        if( slave_sdb->primitive == CON )
        {
          OUTPUT_STRING( "dps_get_status.con(+)" );
        }
        else
        {
          OUTPUT_STRING( "\ndps_get_status.ind" );
        }

        show_get_status_data( &slave_con_ind->dps_get_status_con_ind );
      }

      break;
    } /* case DPS_GET_STATUS */

    /*----------------------------------------------------------------------*/

    default:
    {
      assert( PB_FALSE );
      break;
    } /* default */
  } /* switch */

  return;

} /* show_service */

/****************************************************************************/

FUNCTION PUBLIC VOID get_status_request
  (
  IN    VOID
  )
  /*------------------------------------------------------------------------*/
  /* FUNCTIONAL_DESCRIPTION                                                 */
  /*------------------------------------------------------------------------*/
  /* -This function sends a get_status request.                             */
  /*------------------------------------------------------------------------*/
{
LOCAL_VARIABLES

  T_PROFI_SERVICE_DESCR slave_sdb;

  INT16                 ret_val;

FUNCTION_BODY

  /*--- DPS_GET_STATUS SERVICE DESCRIPTION BLOCK ---------------------------*/

  slave_sdb.comm_ref  = DPS_DEFAULT_COMM_REF;
  slave_sdb.layer     = DPS;
  slave_sdb.service   = DPS_GET_STATUS;
  slave_sdb.primitive = REQ;
  slave_sdb.invoke_id = DPS_DEFAULT_INVOKE_ID;
  slave_sdb.result    = POS;

  OUTPUT_STRING( "dps_get_status.req" );
  ret_val = profi_snd_req_res( &slave_sdb, NULL, 0 );

  if( ret_val != E_OK)
  {
    get_interface_error_string( ret_val, out_string, sizeof( out_string ) );
    OUTPUT_STRING( out_string );
  }

  return;

} /* get_status_request */

/****************************************************************************/
/*** INITIALIZE / TERMINATE DP-SLAVE ****************************************/
/****************************************************************************/

FUNCTION PUBLIC PB_BOOL slave_init_request
  (
  IN    USIGN8    slave_add
  )
  /*------------------------------------------------------------------------*/
  /* FUNCTIONAL_DESCRIPTION                                                 */
  /*------------------------------------------------------------------------*/
  /* -The slave_add is the current address of the station below this station*/
  /*  sends and receives telegrams.                                         */
  /* -The parameter min_tsdr is the minimum waiting time for a DP-Slave     */
  /*  until it is allowed to send the response frames to the DP-Master.     */
  /* -The parameters auto_cfg_response, auto_prm_response and               */
  /*  auto_startup_inputs determine whether the user must acknowlegde the   */
  /*  indications of set_prm and chk_cfg and set input data befor the slave */
  /*  can change the status to data_exchange.                               */
  /* -The parameters xxx_supported support the possibility of synchronize   */
  /*  the output data, freeze the input data and set a new slave address    */
  /*  through the DP master.                                                */
  /* -The parameters xxx_max_len determine the size of the several buffer in*/
  /*  the DPRAM between C165 and ASIC SPC3. The confirmation contains the   */
  /*  number of free ressources or the number of deficiece ressources. The  */
  /*  default values use the available memory. The max_cfg-, max_usr_prm-,  */
  /*  max_ext_diag- and max_address_data_len is set to the default values   */
  /*  specified by the DP standard (EN 50170 or DIN 19245-3). The max_input-*/
  /*  and max_output_data_len use the rest of memory.                       */
  /* -The DP-Slave accepts only parameter frames if the transmitted         */
  /*  ident_number is identical with the own ident_number, which is         */
  /*  established by PNO.                                                   */
  /* -The parameter user_watchdog_timeout is disabled. Therefore the        */
  /*  firmware don't control the grip on the data interface by the          */
  /*  application.                                                          */
  /*------------------------------------------------------------------------*/
{
LOCAL_VARIABLES

  T_PROFI_SERVICE_DESCR  init_slave_sdb;
  T_DPS_INIT_SLAVE_REQ   init_slave_req;

  INT16                  ret_val;

  USIGN8                 input_data_length;
  USIGN8                 output_data_length;


FUNCTION_BODY

  memset(&init_slave_req,0,sizeof(T_DPS_INIT_SLAVE_REQ));

  /*--- DPS_INIT_SLAVE INPUT DATA LEN --------------------------------------*/

  ret_val = convert_cfg_to_length( DPS_DEFAULT_CFG_DATA,
                                   sizeof( DPS_DEFAULT_CFG_DATA ),
                                   &input_data_length,
                                   &output_data_length );

  if( !ret_val )
  {
    input_length  = input_data_length;

    /*--- DPS_INIT_SLAVE SERVICE DESCRIPTION BLOCK -------------------------*/

    init_slave_sdb.comm_ref                = DPS_DEFAULT_COMM_REF;
    init_slave_sdb.layer                   = DPS;
    init_slave_sdb.service                 = DPS_INIT_SLAVE;
    init_slave_sdb.primitive               = REQ;
    init_slave_sdb.invoke_id               = DPS_DEFAULT_INVOKE_ID;
    init_slave_sdb.result                  = POS;

    /*--- BASIC INITIALIZE SLAVE REQUEST -----------------------------------*/

    init_slave_req.slave_add               = slave_add;
    init_slave_req.min_tsdr                = DPS_DEFAULT_MIN_TSDR;

    init_slave_req.auto_cfg_response       = PB_TRUE;
    init_slave_req.auto_prm_response       = PB_TRUE;
    init_slave_req.auto_startup_inputs     = PB_FALSE;

    init_slave_req.sync_mode_supported     = PB_TRUE;
    init_slave_req.freeze_mode_supported   = PB_TRUE;
    init_slave_req.set_slave_add_supported = PB_TRUE;

    init_slave_req.max_input_data_len      = DPS_DEFAULT_INPUT_LEN;
    init_slave_req.max_output_data_len     = DPS_DEFAULT_OUTPUT_LEN;
    init_slave_req.max_cfg_data_len        = DPS_DEFAULT_CFG_DATA_LEN;
    init_slave_req.max_usr_prm_data_len    = DPS_DEFAULT_USR_PRM_DATA_LEN;
    init_slave_req.max_ext_diag_data_len   = DPS_DEFAULT_EXT_DIAG_DATA_LEN;
    init_slave_req.max_address_data_len    = DPS_DEFAULT_ADDRESS_DATA_LEN;

    init_slave_req.ident_number            = DPS_DEFAULT_IDENT_NUMBER;

    init_slave_req.user_watchdog_timeout   = DPS_USER_WATCHDOG_DISABLED;

    init_slave_req.cfg_data_len            = sizeof( DPS_DEFAULT_CFG_DATA );
    init_slave_req.enhanced_init_data_len  = 0;

    /*--- SLAVE CONFIGURATION ----------------------------------------------*/

    memcpy( &init_slave_req.cfg_data, DPS_DEFAULT_CFG_DATA, sizeof( DPS_DEFAULT_CFG_DATA ) );

    /*--- INITIALIZE SLAVE SEND REQUEST ------------------------------------*/

    OUTPUT_STRING( "dps_init_slave" );
    ret_val = profi_snd_req_res( &init_slave_sdb, &init_slave_req, 0 );

    /*--- GET AND DISPLAY INTERFACE ERROR DESCRIPTION ----------------------*/

    if( ret_val != E_OK)
    {
      get_interface_error_string( ret_val, out_string, sizeof( out_string ) );
      OUTPUT_STRING( out_string );
      return( PB_FALSE );
    }

    return( PB_TRUE );

  } /* if valid cfg data */
  else
  {
    get_interface_error_string( ret_val, out_string, sizeof( out_string ) );
    OUTPUT_STRING( out_string );
    return( PB_FALSE );

  } /* else valid cfg data */

} /* slave_init_request */

/****************************************************************************/

FUNCTION PUBLIC VOID slave_exit_request
  (
  IN    VOID
  )
  /*------------------------------------------------------------------------*/
  /* FUNCTIONAL_DESCRIPTION                                                 */
  /*------------------------------------------------------------------------*/
  /* - This function terminate the dp-slave-firmware on PROFI104/PROFIboard */
  /*------------------------------------------------------------------------*/
{
LOCAL_VARIABLES

  T_PROFI_SERVICE_DESCR exit_slave_sdb;

  INT16                 ret_val;

FUNCTION_BODY

  /*--- DPS_EXIT_SLAVE SERVICE DESCRIPTION BLOCK ---------------------------*/

  exit_slave_sdb.comm_ref  = DPS_DEFAULT_COMM_REF;
  exit_slave_sdb.layer     = DPS;
  exit_slave_sdb.service   = DPS_EXIT_SLAVE;
  exit_slave_sdb.primitive = REQ;
  exit_slave_sdb.invoke_id = DPS_DEFAULT_INVOKE_ID;
  exit_slave_sdb.result    = POS;

  /*--- TERMINATE SLAVE SEND REQUEST ---------------------------------------*/

  OUTPUT_STRING( "dps_exit_slave" );
  ret_val = profi_snd_req_res( &exit_slave_sdb, NULL, 0 );

  /*--- GET AND DISPLAY INTERFACE ERROR DESCRIPTION ------------------------*/

  if( ret_val != E_OK)
  {
    get_interface_error_string( ret_val, out_string, sizeof( out_string ) );
    OUTPUT_STRING( out_string );
  }

  return;

} /* slave_exit_request */

/****************************************************************************/
/*** INITIALIZATION / TERMINATION  PAPI AND PROFIBUS CONTROLLER *************/
/****************************************************************************/

FUNCTION LOCAL PB_BOOL hardware_init
  (
  IN    USIGN8    board
  )
  /*------------------------------------------------------------------------*/
  /* FUNCTIONAL_DESCRIPTION                                                 */
  /*------------------------------------------------------------------------*/
  /* Initialize the profibus-controller and the profibus-interface. This    */
  /* function is called before any communication with the profibus-hardware */
  /* is started.                                                            */
  /*------------------------------------------------------------------------*/
{
LOCAL_VARIABLES

  CSTRING     papi_version[ MAX_ERROR_DESCR_LEN ];
  CSTRING     firmware_version[ MAX_ERROR_DESCR_LEN ];

  INT16       ret_val;

FUNCTION_BODY

  OUTPUT_STRING( "initalize Board and CMI" );

  /*--- INITIALIZE PAPI AND START PROFIBUS CONTROLLER ----------------------*/

  if( E_OK != ( ret_val = profi_init(board,0,0) ) )
  {

    /*--- GET AND DISPLAY ERROR DESCRIPTION --------------------------------*/

    get_interface_error_string( ret_val, out_string, sizeof( out_string ) );
    OUTPUT_STRING( out_string );
    return( PB_FALSE );

  }

  /*--- GET PAPI VERSION NUMBER AND FIRMWARE VERSION NUMBER ----------------*/

  (VOID) profi_get_versions( papi_version, firmware_version );

  OUTPUT_STRING( papi_version );
  OUTPUT_STRING( firmware_version );

  return( PB_TRUE );

} /* hardware_init */

/****************************************************************************/

FUNCTION LOCAL VOID hardware_exit
  (
  IN  VOID
  )
  /*------------------------------------------------------------------------*/
  /* FUNCTIONAL_DESCRIPTION                                                 */
  /*------------------------------------------------------------------------*/
  /* Shut down the profibus-controller and the profibus-interface. This     */
  /* function is called to close any communication with the profibus-       */
  /* hardware.                                                              */
  /*------------------------------------------------------------------------*/
{
LOCAL_VARIABLES

FUNCTION_BODY

  (VOID) profi_end( );

  return;

} /* hardware_exit */

/****************************************************************************/
/*** INTERRUPT SERVICE ROUTINES *********************************************/
/****************************************************************************/


#if defined (WIN32) || defined (_WIN32) || defined (WIN16) || defined (_WIN16)
#pragma check_stack (off)
FUNCTION VOID CALLBACK timer_proc
  (
  IN UINT   uID,
  IN UINT   uMsg,
  IN DWORD  dwUser,
  IN DWORD  dw1,
  IN DWORD  dw2
  )
  /*------------------------------------------------------------------------*/
  /* FUNCTIONAL_DESCRIPTION                                                 */
  /*------------------------------------------------------------------------*/
  /* - timer interrupt service routine chained with the DOS system timer    */
  /*   interrupt (55ms)                                                     */
  /* - The interrupt will be acknowledged automatically by the system timer */
  /*   interrupt handler. Thus this function calls the system ISR by means  */
  /*   of the 'chain_intr' routine.                                         */
  /* - In case of using user defined interrupt handler (e.g. peripheral     */
  /*   devices or boards) the respective interrupt has to be acknowleged    */
  /*   in the interrupt controller. Use macro 'outp (EOI_REG, EOI)' for     */
  /*   that.                                                                */
  /*------------------------------------------------------------------------*/
{
LOCAL_VARIABLES

FUNCTION_BODY

  /*--- SIMULATE NEW INPUT DATA --------------------------------------------*/

  dpssched_send_event( DPSSCHED_INPUT_IF_TASK_ID, DPSSCHED_NEW_INPUT_EVENT );

  actual_time++;

  return;

} /* timer_proc */
#pragma check_stack
#endif

/****************************************************************************/
/*** CALLBACK FUNCTIONS *****************************************************/
/****************************************************************************/

FUNCTION GLOBAL VOID papi_task
  (
    IN T_DPSSCHED_EVENT_ID    event_id
  )
  /*------------------------------------------------------------------------*/
  /* FUNCTIONAL_DESCRIPTION                                                 */
  /*------------------------------------------------------------------------*/
  /* -This callback function is called by the scheduler in polling mode and */
  /*  accept only a DPSSCHED_SCHEDULE_EVENT. All other events are illegal.  */
  /*  During every polling cycle function handles the papi interface        */
  /*  services and call the display function if a confirmation, indication  */
  /*  or error received from interface.                                     */
  /*------------------------------------------------------------------------*/
{
LOCAL_VARIABLES

  T_PROFI_SERVICE_DESCR  slave_sdb;
  T_DPSPROFI_CON_IND     slave_con_ind;

  USIGN16                con_ind_len;

  INT16                  ret_val;


FUNCTION_BODY

  switch( event_id )
  {
    case DPSSCHED_SCHEDULE_EVENT:
    {

      profi_snd_idle();

      con_ind_len = sizeof( T_DPSPROFI_CON_IND );
      ret_val     = profi_rcv_con_ind( &slave_sdb, &slave_con_ind, &con_ind_len );

      Sleep(1);

      switch (ret_val)
      {
        case NO_CON_IND_RECEIVED: break;

        /*------------------------------------------------------------------*/

        case CON_IND_RECEIVED:
        {
          show_service( &slave_sdb, &slave_con_ind );
          break;
        }

        /*------------------------------------------------------------------*/

        case E_IF_FATAL_ERROR:
        {
          get_interface_error_string( ret_val, out_string, sizeof( out_string ) );
          OUTPUT_STRING( out_string );
          get_fatal_error_string( (T_EXCEPTION FAR*) &slave_con_ind, out_string, sizeof( out_string ) );
          OUTPUT_STRING( out_string );
          break;
        }

        /*------------------------------------------------------------------*/

        case E_IF_INVALID_DATA_SIZE:
        {
          get_layer_service_string( slave_sdb.layer, slave_sdb.service, out_string, sizeof( out_string ) );
          OUTPUT_STRING( out_string );
          sprintf( out_string, "\tcomm_ref  : 0x%04X ", slave_sdb.comm_ref );
          OUTPUT_STRING( out_string );
          sprintf( out_string, "\tinvoke_id : 0x%04X ", slave_sdb.invoke_id );
          OUTPUT_STRING( out_string );
          break;
        }

        /*------------------------------------------------------------------*/

        default:
        {
          get_interface_error_string( ret_val, out_string, sizeof( out_string ) );
          OUTPUT_STRING( out_string );
          break;
        }

      } /* switch ret_val */

      break;
    } /* case DPSSCHED_SCHEDULE_EVENT */

    /*------------------------------------------------------------------------*/

    case DPSSCHED_GET_STATUS_EVENT:
    {
      get_status_request( );
      break;
    } /* case DPSSCHED_GET_STATUS_EVENT */

    /*------------------------------------------------------------------------*/

    default:
    {
      assert (PB_FALSE);
      break;
    }

  } /* switch event_id */

  return;

} /* papi_task */

/****************************************************************************/

FUNCTION GLOBAL VOID input_task
  (
    IN T_DPSSCHED_EVENT_ID    event_id
  )
  /*------------------------------------------------------------------------*/
  /* FUNCTIONAL_DESCRIPTION                                                 */
  /*------------------------------------------------------------------------*/
  /* - This Callback Function handles the input data.                       */
  /* - If the interrupt service routine generate the event                  */
  /*   DPSSCHED_NEW_INPUT_EVENT this function increments a counter to       */
  /*   simulate asynchron new input data and send it.                       */
  /* - In the other case after the event DPSSCHED_SCHEDULE_EVENT this       */
  /*   function read the value of BIOS-timer to set new input data.         */
  /* - All other events are illegal.                                        */
  /*------------------------------------------------------------------------*/
{
LOCAL_VARIABLES

  USIGN8     state;

  INT16      ret_val;

  USIGN32    current_time;

FUNCTION_BODY

  switch( event_id )
  {
    case DPSSCHED_NEW_INPUT_EVENT :
    {
      input_data[ 0 ]++;

      ret_val = profi_set_dps_input_data( input_data, input_length, &state);

//      sprintf( out_string, "DPSSCHED_NEW_INPUT_EVENT inputs written: input_state = 0x%02hX", state);
//      OUTPUT_STRING( out_string );
//      show_hex_dump( input_data, 5);



      if( ret_val != E_OK )
      {
        get_interface_error_string( ret_val, out_string, sizeof( out_string ) );
        OUTPUT_STRING( out_string );
      }
      break;
    } /* case DPSSCHED_NEW_INPUT_EVENT */

    /*----------------------------------------------------------------------*/

    case DPSSCHED_SCHEDULE_EVENT :
    {
      current_time    = actual_time;
      input_data[ 1 ] = (USIGN8) current_time;
      input_data[ 2 ] = (USIGN8) ( current_time >> 8  );
      input_data[ 3 ] = (USIGN8) ( current_time >> 16 );
      input_data[ 4 ] = (USIGN8) ( current_time >> 24 );

      ret_val = profi_set_dps_input_data( input_data, input_length, &state);

  //    sprintf( out_string, "DPSSCHED_SCHEDULE_EVENT inputs written: input_state = 0x%02hX", state);
  //    OUTPUT_STRING( out_string );
  //    show_hex_dump( input_data, 5);

      if( ret_val != E_OK )
      {
        get_interface_error_string( ret_val, out_string, sizeof( out_string ) );
        OUTPUT_STRING( out_string );
      }
      break;
    } /* case DPSSCHED_SCHEDULE_EVENT */

    /*----------------------------------------------------------------------*/

    default:
    {
      assert (PB_FALSE);
      break;
    }

  } /* switch */

  return;

} /* input_task */

/****************************************************************************/

FUNCTION GLOBAL VOID output_task
  (
    IN T_DPSSCHED_EVENT_ID    event_id
  )
  /*------------------------------------------------------------------------*/
  /* FUNCTIONAL_DESCRIPTION                                                 */
  /*------------------------------------------------------------------------*/
  /* - This Callback Function handles the display of the output data, if a  */
  /*   timer event was received.                                            */
  /* - All other events are illegal.                                        */
  /*------------------------------------------------------------------------*/
{
LOCAL_VARIABLES

  USIGN8      state;
  USIGN8      output_length;
  INT16       ret_val;

FUNCTION_BODY

  switch( event_id )
  {
    case DPSSCHED_TIMER_EVENT:
    {
      output_length = DP_MAX_OUTPUT_DATA_LEN;
      ret_val       = profi_get_dps_output_data( output_data, &output_length, &state);
      if( ret_val == E_OK )
      {
        sprintf( out_string, "reading outputs: output_state = 0x%02hX ",state);
        OUTPUT_STRING( out_string );
        show_hex_dump( output_data, output_length );
      }
      else
      {
        get_interface_error_string( ret_val, out_string, sizeof( out_string ) );
        OUTPUT_STRING( out_string );
      }
      break;
    } /* case DPSSCHED_TIMER_EVENT */

    /*----------------------------------------------------------------------*/

    default:
    {
      assert( PB_FALSE );
      break;
    }

  } /* switch */

  return;

} /* output_task */

/****************************************************************************/
/*** INIT / EXIT MODULE *****************************************************/
/****************************************************************************/

FUNCTION PUBLIC PB_BOOL dpsprofi_init
  (
  IN    USIGN8          board,
  IN    USIGN8          slave_add
  )
  /*------------------------------------------------------------------------*/
  /* FUNCTIONAL_DESCRIPTION                                                 */
  /*------------------------------------------------------------------------*/
  /* - initializes this module                                              */
  /*------------------------------------------------------------------------*/
{
LOCAL_VARIABLES

   MMRESULT mmresult;

FUNCTION_BODY

  /*--- LOCAL VARIABLES ----------------------------------------------------*/

  out_string [0]  = '\0';

  memset( input_data, 0, sizeof( input_data ) );
  memset( output_data, 0, sizeof( output_data ) );

  /*--- INITIALIZE PAPI AND START PROFIBUS CONTROLLER ----------------------*/

  if(! hardware_init( board ) )
  {
    return( PB_FALSE );
  }

  /*--- INITIALIZE DP SLAVE ------------------------------------------------*/

  if(! slave_init_request( slave_add ) )
  {
    return( PB_FALSE );
  }

  /*--- CREATE NEW TASKS (PAPI POLLING, INPUT DATA, OUTPUT DATA) -----------*/

  dpssched_create_task( DPSSCHED_SERVICE_IF_TASK_ID, DPSSCHED_PRIORITY_5, papi_task );
  dpssched_create_task( DPSSCHED_INPUT_IF_TASK_ID, DPSSCHED_PRIORITY_10, input_task );
  dpssched_create_task( DPSSCHED_OUTPUT_IF_TASK_ID, DPSSCHED_NO_SCHEDULE, output_task );

  /*--- INITIALIZE A CYCLIC TIMER WITH DURATION OF 1 SEC ---------------------*/

  dpssched_start_timer(DPSSCHED_OUTPUT_IF_TASK_ID, DPS_CYCLIC_OUTPUT_TIMER_DURATION, PB_TRUE );



  /* --- START 55ms TIMER ------------------------------------------------- */
  mmresult = timeSetEvent(55,55,(LPTIMECALLBACK) timer_proc,0,TIME_PERIODIC);
  dpsprofi_timer_id = (UINT) mmresult;

  if   (mmresult) return(PB_TRUE);
  else            return(PB_FALSE);


} /* dpsprofi_init */

/****************************************************************************/

FUNCTION PUBLIC VOID dpsprofi_exit
  (
  IN    VOID
  )
  /*------------------------------------------------------------------------*/
  /* FUNCTIONAL_DESCRIPTION                                                 */
  /*------------------------------------------------------------------------*/
  /* - terminates this module                                               */
  /*------------------------------------------------------------------------*/
{
LOCAL_VARIABLES

FUNCTION_BODY

  dpssched_stop_timer( DPSSCHED_OUTPUT_IF_TASK_ID );

  /*--- KILL DPSSPROFI TIMER ------------------------------------------------*/
  timeKillEvent(dpsprofi_timer_id);


  /*--- TERMINATE PAPI AND START PROFIBUS CONTROLLER -----------------------*/

  hardware_exit( );

  return;

} /* dpsprofi_exit */

/*** END OF DPSProfi.c ******************************************************/
